from blues_lib.behavior.Trigger import Trigger

class UrlChanges(Trigger):

  def _trigger(self)->bool:
    '''
    if the url changes in the wait time, it will ignore the query params and fragments
    @returns {bool}
    '''
    url = self._config.get('url')
    wait_time = self._config.get('wait_time',3)
    return self._browser.waiter.ec.url_changes(url,wait_time)



from blues_lib.behavior.Bean import Bean

class Html(Bean):

  def _get(self)->str:
    selector:str = self._config.get('loc_or_elem')
    parent_selector:str = self._config.get('parent_loc_or_elem')
    return self._browser.script.javascript.html(selector,parent_selector)

  def _set(self):
    selector:str = self._config.get('loc_or_elem')
    parent_selector:str = self._config.get('parent_loc_or_elem')
    value:str = self._config.get('value','')
    if self._to_be_presence():
      return self._browser.script.javascript.html(selector,value,parent_selector)


from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model
from blues_lib.behavior.factory.BhvFactory import BhvFactory

class BhvExecutor():

  def __init__(self,model:Model,browser=None):
    self._model = model
    self._browser = browser
    
  def execute(self)->STDOut: 
    try:
      value = self._execute(self._model.config,self._model.meta)
      return STDOut(200,'ok',value)
    except Exception as e:
      return STDOut(500,str(e),None)

  def _execute(self,config:any,meta:any)->any:
    
    if isinstance(config,dict):
      kind = config.get('_kind')
      if kind:
        # skip the optional behavior
        if config.get('_skip'):
          return None
        return self._execute_bhv(meta,kind)

      return self._execute_map(config,meta)

    if isinstance(config,list):
      return self._execute_list(config,meta)

    return config
    
  def _execute_bhv(self,meta:dict,kind:str)->any:
    # must pass the meta and bizdata
    model = Model(meta,self._model.bizdata)
    bhv = BhvFactory(model,self._browser).create(kind)
    if not bhv:
      # lazy to load the class
      from blues_lib.behavior.factory.BhvUnitFactory import BhvUnitFactory
      bhv = BhvUnitFactory(model,self._browser).create(kind)
      if not bhv:
        return None

    stdout = bhv.execute()
    if stdout.code != 200:
      raise Exception(stdout.message)
    return stdout.data
  
  def _execute_list(self,configs:list[dict],meta:list[dict])->list[any]:
    values:list[any] = []
    for idx,config in enumerate(configs):
      value = self._execute(config,meta[idx])
      values.append(value)
    return values

  def _execute_map(self,configs:dict[str,dict],meta:dict[str,dict])->dict[str,any]:
    values:dict[str,any] = {}
    for key,config in configs.items():
      value = self._execute(config,meta[key])
      values[key] = value
    return values



import re
import os
from blues_lib.hocon.replacer.BaseReplacer import BaseReplacer

class EnvReplacer(BaseReplacer):

  '''
  replace by the environment variable
  "${ENV_VAR}" 
  '''
  VAR_PATTERN = r"^\$\{([a-zA-Z0-9_]+)\}$"

  def _replace_node_var(self, value:str)->str:
    """替换单个字符串中的环境变量占位符（要求完全匹配）"""
    # 检查是否为完整的环境变量
    match = re.match(self.VAR_PATTERN, value)
    if not match:
      return value

    key:str = match.group(1) 
    return os.environ.get(key)



import re

from blues_lib.model.VarReplacer import VarReplacer

class Model:
  def __init__(self, meta: list[dict]|dict, bizdata: dict = None) -> None:
    self.meta = self.get_etl_meta(meta)
    self.bizdata = self.get_etl_bizdata(bizdata)
    self.config = self.interpolate(self.meta, self.bizdata)
    
  def get_etl_meta(self,meta: list[dict]|dict)->list[dict]|dict:
    # just return in the base class
    return meta
    
  def get_etl_bizdata(self,bizdata)->dict:
    return VarReplacer.replace(bizdata)
    
  def refresh(self):
    # the meta or bizdata may be updated by the user
    self.config = self.interpolate(self.meta, self.bizdata)

  def interpolate(self, meta: list[dict]|dict, bizdata: dict) -> list[dict]|dict:
    """
    递归替换meta中的占位符(${key})为bizdata中对应的值，不修改原对象
    
    Args:
      meta: 包含占位符的字典或列表
      bizdata: 用于替换占位符的数据字典
    
    Returns:
      替换后的新结构
    """
    if not bizdata:
      return meta

    if isinstance(meta, dict):
      return self._process_dict(meta, bizdata)
    elif isinstance(meta, list):
      return self._process_list(meta, bizdata)
    else:
      return meta  # 非容器类型直接返回

  def _process_dict(self, obj: dict, data: dict) -> dict:
    """递归处理字典，创建新对象"""
    result = {}
    for key, value in obj.items():
      if isinstance(value, dict):
        result[key] = self._process_dict(value, data)
      elif isinstance(value, list):
        result[key] = self._process_list(value, data)
      elif isinstance(value, str):
        result[key] = self._replace_placeholders(value, data)
      else:
        result[key] = value  # 非字符串类型保持原样
    return result

  def _process_list(self, arr: list, data: dict) -> list:
    """递归处理列表，创建新对象"""
    result = []
    for item in arr:
      if isinstance(item, dict):
        result.append(self._process_dict(item, data))
      elif isinstance(item, list):
        result.append(self._process_list(item, data))
      elif isinstance(item, str):
        result.append(self._replace_placeholders(item, data))
      else:
        result.append(item)  # 非字符串类型保持原样
    return result

  def _replace_placeholders(self, s: str, data: dict) -> str|int|float|bool|list|dict|None:
    """替换字符串中的占位符，保留原始类型"""
    if not isinstance(s, str):
      return s
      
    # 检查整个字符串是否就是一个占位符
    full_match = re.fullmatch(r'\$\{([^}]+)\}', s)
    if full_match:
      expr = full_match.group(1)
      
      # 处理带默认值的情况: ${key:-default}
      if ':-' in expr:
        key, default_str = expr.split(':-', 1)
        value = data.get(key, default_str)
        # 尝试将默认值转换为原始类型
        if isinstance(value, str):
          return self._convert_to_original_type(value)
        return value
      
      # 处理简单占位符: if data has no key, return None
      value = data.get(expr, None)
      return value
    
    # 处理包含多个占位符或其他文本的字符串
    def replace_match(match):
      expr = match.group(1)
      
      # 处理带默认值的情况: ${key:-default}
      if ':-' in expr:
        key, default = expr.split(':-', 1)
        return str(data.get(key, default))
      
      # 处理简单占位符: if data has no key, return ''
      return str(data.get(expr, ''))
    
    return re.sub(r'\$\{([^}]+)\}', replace_match, s)

  def _convert_to_original_type(self, value_str: str) -> str|int|float|bool|list|dict|None:
    """尝试将字符串转换为原始类型"""
    # 处理布尔值
    if value_str.lower() == 'true':
      return True
    if value_str.lower() == 'false':
      return False
    
    # 处理None
    if value_str.lower() == 'none':
      return None
    
    # 处理数字
    try:
      num = int(value_str)
      return num
    except ValueError:
      try:
        num = float(value_str)
        return num
      except ValueError:
        pass
    
    # 处理列表和字典（如果需要更复杂的解析，可以使用ast.literal_eval）
    if value_str.startswith('[') and value_str.endswith(']'):
      # 简单列表解析，不处理嵌套结构
      items = [item.strip() for item in value_str[1:-1].split(',')]
      return items
    
    if value_str.startswith('{') and value_str.endswith('}'):
      # 简单字典解析，不处理嵌套结构
      pairs = [pair.strip() for pair in value_str[1:-1].split(',')]
      result = {}
      for pair in pairs:
        if ':' in pair:
          key, val = pair.split(':', 1)
          result[key.strip()] = val.strip()
      return result
    
    # 默认返回字符串
    return value_str


import sys

class OSystem():

  @classmethod
  def get_os_type(cls):
    is_windows = sys.platform.startswith('win')
    is_linux = sys.platform.startswith('linux')
    is_mac = sys.platform.startswith('darwin')
    if is_windows:
      return 'windows'
    elif is_linux:
      return 'linux'
    elif is_mac:
      return 'mac'
    else:
      return 'unknown'
 


import json

from blues_lib.type.output.SQLSTDOut import SQLSTDOut
from blues_lib.dao.sql.BluesSQLIO import BluesSQLIO

class TableMutator():

  _SQL_IO = None

  def __init__(self,table:str,should_encode:bool=True)->None:
    self._table = table
    self._should_encode = should_encode

  @property
  def _io(self)->BluesSQLIO:
    if not self._SQL_IO:
      self._SQL_IO = BluesSQLIO()
    return self._SQL_IO
  
  def post(self,fields:list[str],values:list[list[any]])->SQLSTDOut:
    '''
    Insert one more multi rows
    Parameter:
      fields {list<str>} : the table fields 
      values {list<list<str>>} : two-dim list, every list is a row field value
        the fields' length must be equal to the values row list's length
    Return:
      {dict} : the standard sql output,like:
         {'code': 200, 'count': 1, 'sql': 'insert xxx'}
    '''
    self._encode_entities(values)
    return self._io.post(self._table,fields,values)
  
  def insert(self,entities:list[dict])->SQLSTDOut:
    '''
    Insert one more multi rows
    Parameter:
      entities {dict | list<dict>} : every dict contains field:value
    Return:
      {dict} : the standard sql output,like:
         {'code': 200, 'count': 1, 'sql': 'insert xxx'}
    '''
    self._encode_entities(entities)
    return self._io.insert(self._table,entities)

  def put(self,fields:list[str],values:list[any],conditions:dict|list[dict])->SQLSTDOut:
    '''
    Update one or multi rows
    Parameter:
      fields {list<str>} : the table fields 
      values {list<any>} : the updated value
        the value's list length should be equal to the field's list length
      conditions {dict | list<dict>} : the standard condition value ,like:
        {'field':'material_id','comparator':'=','value':'id2'} 
        [{'field':'material_id','comparator':'=','value':'id2'}]
    Return:
      {dict} : the standard sql output,like:
         {'code': 200, 'count': 1, 'sql': 'update xxx'}
    '''
    self._encode_entity(values)
    return self._io.put(self._table,fields,values,conditions)

  def update(self,entity:dict,conditions:dict|list[dict])->SQLSTDOut:
    '''
    Update one or multi rows
    Parameter:
      entity {dict} : every dict contains field:value
      conditions {dict | list<dict>} : the standard condition value ,like:
        {'field':'material_id','comparator':'=','value':'id2'} 
        [{'field':'material_id','comparator':'=','value':'id2'}]
    Return:
      {dict} : the standard sql output,like:
         {'code': 200, 'count': 1, 'sql': 'update xxx'}
    '''
    self._encode_entity(entity)
    return self._io.update(self._table,entity,conditions)

  def delete(self,conditions:dict|list[dict])->SQLSTDOut:
    '''
    Delete one or multi rows
    Parameter:
      conditions {dict | list<dict>} : the standard condition value ,like:
        {'field':'material_id','comparator':'=','value':'id2'} 
        [{'field':'material_id','comparator':'=','value':'id2'}]
    Return:
      {dict} : the standard sql output,like:
         {'code': 200, 'count': 1, 'sql': 'delete xxx'}
    '''
    return self._io.delete(self._table,conditions)
  
  def _encode_entities(self,entities:list[any])->None:
    if self._should_encode and isinstance(entities,list):
      for entity in entities:
        self._encode_entity(entity)

  def _encode_entity(self,entity:dict|list[any])->None:
    if self._should_encode:
      self._encode(entity)

  def _encode(self, entity: dict|list[any]) -> None:
    """
    Encodes non-MySQL-native types in a dictionary or list to JSON strings (single level).
    Ignores inputs that are not dictionaries or lists.
    
    Args:
        entity: A dictionary or list representing data to be written to the database.
    
    Returns:
        The encoded entity (modified in-place for dictionaries, new list for lists),
        or the original input if it's not a dictionary or list.
    """
    mysql_types = {int, float, str, bool, type(None)}
    
    if isinstance(entity, dict):
      for key, value in entity.items():
        if type(value) in mysql_types:
          continue

        try:
          entity[key] = json.dumps(value, ensure_ascii=False)
        except json.JSONDecodeError as e:
          entity[key] = str(e)

    if isinstance(entity, list):
      for index,item in enumerate(entity):
        if type(item) in mysql_types:
          continue

        try:
          entity[index] = json.dumps(item, ensure_ascii=False)
        except json.JSONDecodeError as e:
          entity[index] = str(e)



import subprocess
from blues_lib.util.BluesConsole import BluesConsole
import subprocess

class BluesPowerShell():

  @classmethod
  def execute(cls,script):
    '''
    @description : execute ps script
    @param {str} script str or ps1 file
    @returns {dict} PSResult
    '''
    ps_script = 'powershell.exe %s' % script
    p = subprocess.Popen(ps_script,stdout=subprocess.PIPE,stderr=None)
    output,error = p.communicate()
    try:
      return {
        'code':200,
        'message':'success',
        'output':output.decode('utf-8').strip()
      }
    except:
      return {
        'code':500,
        'message':error,
        'output':output
      }

  @classmethod
  def stop_process(cls,name):
    ps_script = 'Get-Process -Name %s -ErrorAction SilentlyContinue | Stop-Process -Force' % name
    return cls.execute(ps_script)

  @classmethod
  def start_process(cls,file_path,args):
    '''
    @description : start a process, the file path and args must be warppered by single quotes
    @param {str} file_path : the absolute path of the start file (.exe)
    @param {str}  args : the multi exe executor's arguments, split by space
     - '--arg1 --arg2 --arg3'
    '''
    args_list = ''
    if args:
      args_list = '-ArgumentList \'%s\''  % args
    ps_script = 'Start-Process -FilePath \'%s\' %s' % (file_path,args_list)
    BluesConsole.info(ps_script,'Powershell: ')
    return cls.execute(ps_script )

  @classmethod
  def get_env_value(cls,name):
    '''
    @description : get env variable's value
    @param {str} name : varibale's name
    @returns {str} : return var's value or ''
    '''
    ps_script = "[Environment]::GetEnvironmentVariable('%s','machine')" % name
    result = cls.execute(ps_script)
    if result['code'] == 200:
      return result['output']
    else:
      return ''


from blues_lib.type.factory.Factory import Factory
from blues_lib.model.Model import Model
from blues_lib.sele.browser.Browser import Browser 
from blues_lib.crawler.page.NextPageCrawler import NextPageCrawler
from blues_lib.namespace.CrawlerName import CrawlerName

class PageCrawlerFactory(Factory):

  _crawlers = {
    NextPageCrawler.NAME:NextPageCrawler,
  }

  def __init__(self,model:Model,browser:Browser) -> None:
    '''
    @param model {Model} : the model of crawler
    @param browser {Browser} : the browser instance to use
    '''
    self._model = model
    self._browser = browser

  def create(self,name:CrawlerName):
    crawler = self._crawlers.get(name)
    if not crawler:
      return None
    return crawler(self._model,self._browser)



from blues_lib.namespace.CrawlerName import CrawlerName
from blues_lib.crawler.Crawler import Crawler

class BaseCrawler(Crawler):

  def _before_crawled(self):
    # main crawler
    self._crawler_meta = self._meta.get(CrawlerName.Field.CRAWLER.value)
    self._crawler_conf = self._conf.get(CrawlerName.Field.CRAWLER.value)




from blues_lib.command.io.InputMapper import InputMapper
from blues_lib.hocon.replacer.HoconReplacer import HoconReplacer
from blues_lib.command.io.IOExcept import IOExcept

class InputHandler:
  
  def __init__(self,task_def:dict,task_biz:dict,ti:any) -> None:
    self._task_def = task_def
    self._task_biz = task_biz
    self._ti = ti

  def handle(self)->tuple[dict,dict]:
    task_biz = self._get_biz(self._task_def,self._task_biz)
    task_def = self._get_def(self._task_def,task_biz)
    return task_def,task_biz

  def _get_biz(self,task_def:dict,task_biz:dict)->dict:
    # Step1: replace the input mapping placeholder with bizdata
    mapping_def:dict = {
      "input":task_def.get('input')
    }
    mapping_def = HoconReplacer(mapping_def,task_biz).format()

    # Step2: add upstream's output to the bizdata 
    InputMapper(mapping_def,task_biz,self._ti).handle()
    
    # Step3: replace the bizdata placeholder (include and function) with the actual value
    return HoconReplacer(task_biz).replace()

  def _get_def(self,task_def:dict,task_biz:dict)->dict:

    # Step1: replace the placeholder in the task_def
    definition:dict = HoconReplacer(task_def).replace()

    # Step2: replace the variable placeholder in the definition, exclude dag and meta
    if definition.get('dag'):
      definition = self._get_format_def(definition,task_biz,'dag')
    elif definition.get('meta'):
      definition = self._get_format_def(definition,task_biz,'meta')
    else:
      definition = self._get_format_def(definition,task_biz)

    # Step3: validate the task def
    IOExcept.validate_task(definition)
    return definition
  
  def _get_format_def(self,task_def:dict,task_biz:dict,key:str='')->dict:
    if key:
      value:dict = task_def.pop(key)
    definition:dict = HoconReplacer(task_def,task_biz).format()
    if key:
      definition[key] = value
    return definition



from .Alert import Alert

class Navi():
 
  def __init__(self,driver):
    self.__driver = driver
    self.__alert = Alert(driver)
      
  def open(self,url):
    '''
    Open a page
    '''
    if self.is_available():
      self.__driver.get(url)
      
  def get(self,url):
    self.open(url)

  def close(self):
    '''
    Close the current window or tab
    If only one window/tab ,the browser will be closed
    '''
    self.__alert.accept()
    if self.is_available():
      self.__driver.close()

  def is_available(self):
    '''
    Whether the current driver is available
    Returns:
      {bool}
    '''
    try:
      self.__driver.current_url
      return True
    except Exception as e:
      return False

  def quit(self):
    '''
    Close all windows and tabs
    Close the browser process
    Close the driver's process
    '''
    print('------quit------')
    self.__alert.accept()
    if self.is_available():
      self.__driver.quit()

  def back(self):
    '''
    Back to the prev page
    '''
    if self.is_available():
      self.__driver.back()
  
  def forward(self):
    '''
    Forward to the next page
    '''
    if self.is_available():
      self.__driver.forward()

  def refresh(self):
    '''
    Refresh to the current page
    '''
    if self.is_available():
      self.__driver.refresh()
  



from functools import wraps
from .WaiterDeco import WaiterDeco

class QuerierDeco(WaiterDeco):

  caller_class = 'Querier'

  def __call__(self,func):
    @wraps(func) 
    def wrapper(*args,**kwargs):
      return self.wrapper(func,*args,**kwargs)
    return wrapper

  def set_arg_index(self):
    self.arg_cs_index = 1
    self.arg_pcs_index = 2




from blues_lib.model.BaseVarReplacer import BaseVarReplacer
from blues_lib.model.EnvVarReplacer import EnvVarReplacer
from blues_lib.model.CalVarReplacer import CalVarReplacer

class VarReplacer(BaseVarReplacer):
    '''
    Composite variable replacer that combines multiple variable replacers.
    Performs variable replacement using multiple replacers in a single traversal,
    checking each replacer in order until a value is found.
    
    This avoids the need for multiple separate replacements and duplicate traversals.
    '''
    
    # Class-level storage for replacer classes
    _replacer_classes = [EnvVarReplacer, CalVarReplacer]
    
    @classmethod
    def set_replacers(cls, replacer_classes):
        '''
        Set the list of replacer classes to use.
        
        Args:
            replacer_classes: List of replacer classes to use for variable lookup
        '''
        cls._replacer_classes = replacer_classes
    
    @classmethod
    def _get_raw_value(cls, var_name: str):
        '''
        Get value by checking all configured replacers in order.
        
        Args:
            var_name: Variable name to look up
            
        Returns:
            First found variable value from any replacer, or None if not found
        '''
        for replacer_class in cls._replacer_classes:
            value = replacer_class._get_raw_value(var_name)
            if value is not None:
                return value
        return None
    
    @classmethod
    def create_with_replacers(cls, replacer_classes):
        '''
        Create a composite replacer with specified replacers.
        
        Args:
            replacer_classes: List of replacer classes to use
            
        Returns:
            VarReplacer class with configured replacers
        '''
        # Create a new subclass to avoid modifying the original class
        class CustomVarReplacer(cls):
            pass
        CustomVarReplacer._replacer_classes = replacer_classes
        return CustomVarReplacer


import json
from typing import List,Dict,Any

from blues_lib.type.output.SQLSTDOut import SQLSTDOut
from blues_lib.dao.sql.BluesSQLIO import BluesSQLIO    

class TableQuerier():

  _SQL_IO = None

  def __init__(self,table:str,should_decode:bool=True)->None:
    self._table = table
    self._should_decode = should_decode
  
  @property
  def _io(self)->BluesSQLIO:
    if not self._SQL_IO:
      self._SQL_IO = BluesSQLIO()
    return self._SQL_IO
  
  def get(self,fields:List[str]=None,conditions:List[dict]=None,orders:List[dict]=None,pagination:dict=None)->SQLSTDOut:
    '''
    Query rows
    Parameter:
      fields {list<str>} : the table fields 
      conditions {dict | list<dict>} : the standard condition value ,like:
        {'operator':'and','field':'material_id','comparator':'=','value':'id2'} 
        [{'operator':'and','field':'material_id','comparator':'=','value':'id2'}]
      orders {dict | list<dict>} : the standard order by dict, like:
        {'field':'material_status','sort':'asc'}
        [{'field':'material_status','sort':'asc'}]
      pagination {dict} : the standard pager dict, like:
        {'no':1,'size':10}
    Return:
      {SQLSTDOut} : the standard sql output,like:
    '''
    stdout = self._io.get(self._table,fields,conditions,orders,pagination)
    return self._decode_entities(stdout)
  
  def _decode_entities(self,stdout:SQLSTDOut):
    if stdout.code != 200 or not stdout.data or not self._should_decode:
      return stdout

    for entity in stdout.data:
      self._decode(entity)

    return stdout

  def first(self,conditions:List[dict]=None)->SQLSTDOut:
    stdout = self.get(conditions=conditions)
    if stdout.code != 200 or not stdout.data:
      return stdout
    
    return SQLSTDOut(200,'ok',stdout.data[0])

  @classmethod 
  def _decode(cls,entity: Dict[str, Any]) -> None:
    """
    Decodes JSON string fields in a dictionary to Python objects.
    
    This method modifies the input dictionary in-place, converting 
    JSON-formatted string values into their corresponding Python objects.
    Empty or whitespace-only strings are skipped to avoid invalid parsing.
    
    Args:
      entity: A dictionary representing a database row, 
        where keys are field names and values are field values.
    """
    for key, value in entity.items():
      if isinstance(value, str) and value.strip():  
        try:
          entity[key] = json.loads(value)
        except Exception as e:
          continue


from blues_lib.behavior.Bean import Bean

class Screenshot(Bean):

  def _get(self)->any:
    if self._config.get('loc_or_elem'):
      kwargs = self._get_kwargs(['loc_or_elem','file','parent_loc_or_elem','timeout'])
      if self._to_be_presence():
        return self._browser.element.shot.screenshot(**kwargs)
    else:
      file = self._config.get('file','')
      return self._browser.interactor.window.screenshot(file)



from blues_lib.sele.script.javascript.JavaScript import JavaScript
from blues_lib.sele.script.shell.Shell import Shell
from blues_lib.sele.script.python.Python import Python

class Script():

  def __init__(self,driver):
    self.javascript = JavaScript(driver)
    self.shell = Shell()
    self.python = Python(driver)



import re
import re
from bs4 import BeautifulSoup, Comment
import tiktoken

class HtmlExtractor:
  
  def extract(self,html:str, includes:list[str]=[], excludes:list[str]=[])->dict[str,str|int]:
    """
    提取html的body有效内容并移除非必要属性
    这是唯一暴露的公共方法
    
    params:
     - html: 要提取的html字符串
     - includes: 可选的CSS选择器列表，用于在body之后进一步过滤内容，支持多个选择器
     - excludes: 可选的CSS选择器列表，用于排除指定的元素，在includes计算后立即执行，支持多个选择器
    
    returns: dict
     - html: 解析后的html
     - char_size： 解析后的html长度
     - token_size： 解析后的html token长度
    """
    if not html or not isinstance(html, str):
      return {"html": "", "char_size": 0, "token_size": 0}
    
    # 串行调用其他方法
    html = self._body(html, includes, excludes)
    html = self._shake(html)
    html = self._minify(html)
    
    # 计算字符长度
    char_size = len(html)
    
    # 计算token数量 : 会导致程序卡死，待优化
    token_size = -1 # self._count_tokens(html)
    
    return {
      "html": html,
      "char_size": char_size,
      "token_size": token_size
    }
  
  def _count_tokens(self, text: str) -> int:
    """
    计算文本的token数量
    使用tiktoken库进行计算
    """
    try:
      # 使用cl100k_base编码（GPT-4、GPT-3.5-Turbo等使用的编码）
      encoding = tiktoken.get_encoding("cl100k_base")
      tokens = encoding.encode(text)
      return len(tokens)
    except Exception as e:
      # 如果出现异常，返回字符长度作为备选
      return len(text)
  
  def _filter_by_css_selectors(self, soup, css_selectors:list[str])->BeautifulSoup:
    """
    [Protected] 根据CSS选择器列表过滤HTML内容
    
    params:
     - soup: BeautifulSoup对象
     - css_selectors: CSS选择器列表
    
    returns:
     - 过滤后的BeautifulSoup对象
    """
    if not css_selectors:
      return soup
    
    try:
      # 创建一个新的BeautifulSoup对象，用于存储所有选中的元素
      new_soup = BeautifulSoup('<div></div>', 'html.parser')
      container = new_soup.div
      
      # 遍历所有选择器
      for selector in css_selectors:
        if selector:
          selected_elements = soup.select(selector)
          for element in selected_elements:
            # 克隆元素以避免修改原始文档
            container.append(element)
      
      # 如果有选中的元素，返回新的soup对象
      if len(container.contents) > 0:
        return new_soup
    except Exception as e:
      # 如果CSS选择器无效或其他错误，忽略该操作，返回原始soup对象
      pass
    
    return soup
  
  def _exclude_by_css_selectors(self, soup, css_selectors:list[str])->BeautifulSoup:
    """
    [Protected] 根据CSS选择器列表排除HTML内容
    
    params:
     - soup: BeautifulSoup对象
     - css_selectors: CSS选择器列表
    
    returns:
     - 排除指定元素后的BeautifulSoup对象
    """
    if not css_selectors:
      return soup
    
    try:
      # 遍历所有要排除的选择器
      for selector in css_selectors:
        if selector:
          # 查找所有匹配的元素并移除
          for element in soup.select(selector):
            element.decompose()
    except Exception as e:
      # 如果CSS选择器无效或其他错误，忽略该操作，返回原始soup对象
      pass
    
    return soup
    
  def _body(self,html:str, includes:list[str]=[], excludes:list[str]=[])->str:
    """
    [Protected] 提取html的body内容
    - 提取body标签内容
    - 如果提供了includes选择器列表，先使用这些选择器过滤内容
    - 如果提供了excludes选择器列表，在includes过滤后立即执行排除操作
    - 移除指定的不需要的标签：style、script、iframe、frame、canvas、svg
    - 移除所有注释
    """
    if not html or not isinstance(html, str):
      return ""
      
    # 解析HTML
    soup = BeautifulSoup(html, 'html.parser')
    
    # 提取body内容，如果没有body则使用整个文档
    body = soup.body
    if body:
      soup = BeautifulSoup(str(body), 'html.parser')
    
    # 先根据CSS选择器列表过滤内容
    if includes:
      soup = self._filter_by_css_selectors(soup, includes)
    
    # 然后立即执行排除操作
    if excludes:
      soup = self._exclude_by_css_selectors(soup, excludes)
    
    # 定义需要移除的标签列表，方便统一管理和修改
    tags_to_remove = ['style', 'script', 'iframe', 'frame', 'canvas', 'svg', 'footer', 'aside']
    
    # 统一移除所有指定的标签
    for tag_name in tags_to_remove:
      for tag in soup(tag_name):
        tag.decompose()
    
    # 移除所有注释
    for comment in soup.find_all(string=lambda text: isinstance(text, Comment)):
      comment.extract()
    
    # 空白字符的处理将在_minify方法中进行
    return str(soup)
    
  def _shake(self,html:str)->str:
    """
    [Protected] 移除非必要属性，减少总字符数量
    - 移除所有标签内的非必要属性，例如 style class id 等
    - 保留某些必须的属性，例如 href src alt 等
    - 移除样式类标签 strong、b、i，但保留标签内的文本内容
    - 移除base64类型的img标签（src属性以data:image/开头）
    """
    if not html or not isinstance(html, str):
      return ""
      
    # 定义要保留的属性
    keep_attributes = {'href', 'src', 'alt', 'title'}
    
    # 解析HTML
    soup = BeautifulSoup(html, 'html.parser')
    
    # 移除样式类标签 strong、b、i，但保留标签内的文本内容
    style_tags = ['strong', 'b', 'i']
    for tag_name in style_tags:
      for tag in soup.find_all(tag_name):
        tag.unwrap()  # 移除标签但保留内容
    
    # 移除base64类型的img标签
    for img in soup.find_all('img'):
      src = img.get('src', '')
      if src.lower().startswith('data:image/'):
        img.decompose()  # 完全移除该img标签
    
    # 遍历所有标签
    for tag in soup.find_all(True):
      # 创建一个新的字典，只保留必要的属性
      new_attrs = {}
      for attr, value in tag.attrs.items():
        if attr.lower() in keep_attributes:
          new_attrs[attr] = value
      
      # 替换属性
      tag.attrs = new_attrs
    
    # 返回清理后的HTML
    return str(soup)
    
  def _minify(self, html: str) -> str:
    """
    [Protected] 最小化HTML内容
    - 移除所有不必要的空白
    - 移除所有注释
    - 压缩标签间的空白
    """
    if not html or not isinstance(html, str):
      return ""
    
    # 移除注释
    html = re.sub(r'<!--.*?-->', '', html, flags=re.DOTALL)
    # 压缩空白字符
    html = re.sub(r'\s+', ' ', html)
    # 移除标签间的空白
    html = re.sub(r'>\s+<', '><', html)
    # 移除首尾空白
    html = html.strip()
    
    return html


import sys,os,re

from blues_lib.type.output.STDOut import STDOut
from blues_lib.type.chain.Handler import Handler

class AllMatchHandler(Handler):

  def handle(self)->STDOut:
    stdout = self.resolve()

    if self._next_handler:
      return self._next_handler.handle()
    else:
      return stdout




from .ChromeArgs import ChromeArgs

class ChromeExtArgs(ChromeArgs):

  __default_args = [
    # 加载 Chrome 扩展插件,在需要测试扩展功能或加载特定插件时使用。such as  '/path/to/extension.crx'
  ]
    
  def get(self):
    '''
    Get the default and input experimental args
    @returns {list} : removed duplicate args
    '''
    arg_list = self.__default_args+self.get_from_input()
    return list(set(arg_list))
  
  def get_from_input(self):
    '''
    Convert the pass standard args settings to the real args
      - replace the input value to the placehoder
    '''
    args = []
    if not self._input_args:
      return args
    
    for key,value in self._input_args.items():
      if value:
        args.append(key)
    return args



from blues_lib.namespace.CommandName import CommandName
from blues_lib.command.SQLCommand import SQLCommand
from blues_lib.dao.sql.TableQuerier import TableQuerier
from blues_lib.type.output.STDOut import STDOut

class Querier(SQLCommand):

  NAME = CommandName.SQL.QUERIER

  def _invoke(self)->STDOut:
    table = self._summary.get('table')
    conditions = self._summary.get('conditions')
    fields = self._summary.get('fields') or '*'
    pagination = self._summary.get('pagination') or {
      'no':1,
      'size':1
    }
    orders = self._summary.get('orders')
    querier = TableQuerier(table)
    return querier.get(fields,conditions,orders,pagination)



     


import re
from blues_lib.material.MatHandler import MatHandler
from blues_lib.type.output.STDOut import STDOut
from blues_lib.dao.material.MatMutator import MatMutator 

class Sinker(MatHandler):

  def resolve(self)->STDOut:
    self._setup()

    if not self._rule:
      return STDOut(200,'skip this sinker')

    if not self._entities:
      raise Exception(f'{self.__class__.__name__} entities is empty')
    
    self._method = self._rule.get('method','insert')
    conditions:list[dict] = self._rule.get('conditions',[])
    
    sink_entities:list[dict] = self._get_sink_entities()
    
    # insert to the db
    if self._method == 'insert':
      return MatMutator().insert(sink_entities)
    elif self._method == 'update':
      # only update the first entity
      return MatMutator().update(sink_entities[0],conditions)
    elif self._method == 'delete':
      return MatMutator().delete(conditions)
    else:
      return STDOut(500,f'{self.__class__.__name__} no support method: {self._method}')
    
  def _get_sink_entities(self)->list[dict]:
    
    extend_entity:dict = self._rule.get('entity') # the merged fields
    inc_fields:list[str] = self._rule.get('inc_fields',[])
    inc_pattern:str = self._rule.get('inc_pattern','')
    exc_fields:list[str] = self._rule.get('dec_fields',[])
    exc_pattern:str = self._rule.get('dec_pattern','')

    sink_entities:list[dict] = []
    
    for entity in self._entities:
      sink_entity:dict = entity.copy()
      # merge the entity
      if extend_entity:
        sink_entity.update(extend_entity)
        
      # include and exclude the fields
      if inc_fields:
        sink_entity = {k:v for k,v in sink_entity.items() if k in inc_fields}

      if inc_pattern:
        sink_entity = {k:v for k,v in sink_entity.items() if re.match(inc_pattern,k)}

      if exc_fields:
        sink_entity = {k:v for k,v in sink_entity.items() if k not in exc_fields}

      if exc_pattern:
        sink_entity = {k:v for k,v in sink_entity.items() if not re.match(exc_pattern,k)}

      sink_entities.append(sink_entity)
    return sink_entities
    




import os
from blues_lib.util.BluesFiler import BluesFiler
from blues_lib.util.FileReader import FileReader

class DotFileReader:

  @classmethod
  def read_hocon(cls,project_dir:str,dot_path:str):
    '''
    @description: read conf file 
    @param {str} project_dir project root dir
    @param {str} dot_path dot path such as 'tests.mock.command.llm-loop-urls.def'
    @returns {dict|None}
    '''
    file_path = cls.get_file_path(project_dir,dot_path,'conf')
    return FileReader.read_hocon(file_path)

  @classmethod
  def read_text(cls,project_dir:str,dot_path:str):
    file_path = cls.get_file_path(project_dir,dot_path,'txt')
    return FileReader.read_text(file_path)
  
  @classmethod
  def read_json(cls,project_dir:str,dot_path:str):
    file_path = cls.get_file_path(project_dir,dot_path,'json')
    return FileReader.read_json(file_path)
  
  @classmethod
  def read_json(cls,project_dir:str,dot_path:str):
    file_path = cls.get_file_path(project_dir,dot_path,'json')
    return FileReader.read_json(file_path)
  
  @classmethod
  def read_json5(cls,project_dir:str,dot_path:str):
    file_path = cls.get_file_path(project_dir,dot_path,'json5')
    return FileReader.read_json5(file_path)
  
  @classmethod
  def read_yaml(cls,project_dir:str,dot_path:str):
    file_path = cls.get_file_path(project_dir,dot_path,'yaml')
    return FileReader.read_yaml(file_path)
  
  @classmethod
  def read_csv(cls,project_dir:str,dot_path:str):
    file_path = cls.get_file_path(project_dir,dot_path,'csv')
    return FileReader.read_csv(file_path)
  
  @classmethod
  def get_file_path(cls,project_dir:str,dot_path:str,extension:str='')->str:
    file_path = cls.get_slash_path(dot_path,extension)
    return BluesFiler.get_abs_path(project_dir,file_path)

  @classmethod
  def get_slash_path(cls, dot_path: str, extension: str = '') -> str:
    '''
    Convert dot-separated string to file path with optional extension.
    @param dot_path: Dot-separated path (e.g. 'blues_lib.util.BluesFiler' or 'config')
    @param extension: Optional file extension (e.g. 'json' or '.json', both are acceptable)
    @return: File path (e.g. 'blues_lib/util/BluesFiler.json' or 'config.yaml')
    '''
    # Handle empty dot_path
    if not dot_path:
      raise ValueError("dot_path cannot be empty")
    
    # Process extension: unify to ".ext" format (handle with/without leading dot)
    if extension:
      ext = extension if extension.startswith('.') else f'.{extension}'

    if '/' in dot_path or '\\' in dot_path:
      if not dot_path.endswith(ext) and ext:
        dot_path += ext
      return dot_path
     
    # Split dot-separated string into nodes
    path_nodes = dot_path.split('.')
    
    # Append extension to the last node (avoid duplicate extension)
    if not path_nodes[-1].endswith(ext) and ext:
      path_nodes[-1] += ext
    
    # Join nodes with OS-specific path separator
    return os.path.join(*path_nodes)


from .ChromeArgs import ChromeArgs

class ChromeStdArgs(ChromeArgs):

  __default_args = [
    
    # 窗口设置
    '--start-maximized', # 启动浏览器时最大化窗口。在自动化测试中，最大化窗口可以确保页面元素完全可见，避免因分辨率问题导致的元素定位失败。
    
    # 安全设置
    '--ignore-ssl-errors', # 忽略 SSL 证书错误。在测试或爬虫中，访问使用自签名证书的网站时使用。
    '--ignore-certificate-errors', # 忽略证书错误。在测试或爬虫中，访问使用无效证书的网站时使用。
    '--disable-web-security', # 禁用同源策略（用于测试跨域请求）。
    '--allow-running-insecure-content', # 允许运行不安全的内容。
    '--disable-https-first-mode', # 禁用 HTTPS 优先模式。在测试或爬虫中，访问使用 HTTP 协议的网站时使用。
    '--ignore-certificate-errors', # 忽略证书错误。在测试或爬虫中，访问使用无效证书的网站时使用。
    '--disable-features=StrictTransportSecurity', # 禁用严格的传输安全策略。在测试或爬虫中，访问使用 HTTP 协议的网站时使用。
    
    # 工程设置
    '--enable-unsafe-swiftshader', # 启用 SwiftShader（一种软件渲染器）
    '--disable-webgl', # 禁用 WebGL，避免告警
    '--disable-webrtc', # 禁用 WebRTC，避免告警

    # 性能设置
    '--disable-extensions', # 禁用所有 Chrome 扩展程序。在测试中，避免扩展程序干扰浏览器行为。
    '--disable-extensions-api', # 禁用 Chrome 扩展程序的 API。在测试中，避免扩展程序干扰浏览器行为。
    '--disable-popup-blocking', # 禁用浏览器的弹出窗口阻止功能。在测试需要弹出窗口的功能时使用，如广告或登录窗口。
    '--disable-dev-shm-usage', # 禁用 /dev/shm 的使用（适用于 Docker 或内存受限环境）。

    # 自动化相关设置
    '--disable-blink-features=AutomationControlled', # 禁用自动化控制检测（避免被网站识别为自动化工具）。
    '--disable-infobars', # 隐藏 Chrome 顶部的“Chrome 正受到自动测试软件控制”提示栏。在自动化测试中，避免提示栏干扰页面布局或测试逻辑。
    '--enable-clipboard-features=AllowDirectSAP', # 支持剪切板复制 

    # 系统通知检查类设置
    '--disable-notifications', # 禁用浏览器的通知功能。在测试或爬虫中，避免网站弹出通知干扰操作。
    '--hide-crash-restore-bubble', # 隐藏浏览器崩溃后恢复页面的提示气泡。在自动化测试中，避免崩溃恢复提示干扰测试逻辑。
    '--disable-application-install-prompt', # 禁用应用安装提示。在测试中，避免网站提示安装应用干扰测试流程。
    '--no-first-run', # 跳过 Chrome 的首次运行向导。在自动化测试中，避免首次运行向导干扰测试流程。
    '--disable-first-run-ui', # 禁用首次运行时的用户界面（如欢迎页面）。在自动化测试中，避免首次运行界面干扰测试流程。
    '--no-default-browser-check', # 禁止 Chrome 检查是否设置为默认浏览器。在自动化测试中，避免浏览器弹出提示框干扰测试流程。
    '--disable-default-apps', # 禁用默认应用。
    '--disable-geolocation', # 禁用地理位置请求
    '--disable-save-password-bubble', # 禁用保存密码提示
    '--disable-translate', # 禁用翻译提示

    # 环境兼容设置 
    '--disable-gpu', # 禁用 GPU 加速，避免某些环境下的兼容性问题。在无头模式下运行时，某些系统可能会报错，禁用 GPU 可以规避这些问题。
    '--no-sandbox', # 禁用沙盒模式，避免某些环境下的权限问题。在 Linux 或 Docker 环境中运行时，沙盒模式可能会导致浏览器无法启动。
  ]

  # 不再支持字典形式，仅支持列表形式
  __optional_args = {
    'headless':'--headless', # 启用无头模式，浏览器在后台运行，不显示可视化界面。适用于服务器环境或不需要可视化界面的自动化测试场景，如爬虫或持续集成（CI）环境。
    'incognito':'--incognito', # 启用隐身模式（无痕模式），浏览器不会保存历史记录、缓存和 Cookie。 在需要隔离会话的测试场景中，如测试登录功能或避免缓存干扰。
    'imageless':'--blink-settings=imagesEnabled=false', # 禁用图片加载。在爬虫或测试中，禁用图片可以加快页面加载速度。完全禁用所有图片（包括 CSS 背景）
    # 语言
    'lang':'--lang=$()', # 系统语言 such as: zh-CN
    # 启用远程调试端口,Chrome 浏览器的启动参数，用于指定 Chrome 启动时监听的调试端口。 such as: 9222
    # 当你需要启动一个新的 Chrome 浏览器实例，并希望它启用远程调试功能时，才需要设置这个参数。
    'debugport':'--remote-debugging-port=$()', 
    # 指定 Chrome 监听调试连接的网络接口（IP 地址）。默认是本地 127.0.0.1 - 本地机器连接；0.0.0.0 允许任何设备通过 IP 地址和端口连接到 Chrome 的调试服务。
    'debugaddr':'--remote-debugging-address=$()',
    # 用户信息目录
    'debugdir':'--user-data-dir=$()',
    # 日志设置
    'logable':'--enable-logging', # 启用浏览器日志输出。
    'loglevel':'--v=$()', # 设置日志详细级别（1 为最低，9 为最高）such as: 1
  }

  def get(self)->list[str]:
    '''
    Get the default and input experimental args
    @returns {list} : removed duplicate args
    '''
    arg_list = self.__default_args+self.get_from_input()
    return list(set(arg_list))
  
  def get_from_input(self):
    '''
    Convert the pass standard args settings to the real args
      - replace the input value to the placehoder
    '''
    args = []
    if not self._input_args:
      return args

    # 模式1：列表形式，支持所有属性，直接设置规范的--xx属性 
    args.extend(self._input_args)
    return args



from selenium.webdriver.remote.webelement import WebElement
from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model
from blues_lib.namespace.CrawlerName import CrawlerName
from blues_lib.crawler.base.BaseCrawler import BaseCrawler
from blues_lib.hocon.replacer.HoconReplacer import HoconReplacer

class ForCrawler(BaseCrawler):

  NAME = CrawlerName.Engine.FOR

  def _get_entities(self)->list[dict]|None:
    '''
    Set the entities for for crawler
    @return {None}
    '''
    entities:list[dict] = self._summary_conf.get(CrawlerName.Field.ENTITIES.value) 
    if entities:
      return entities

    for_element:str = self._summary_conf.get('for_element')
    if for_element:
      elements:list[WebElement] = self._browser.waiter.querier.query_all(for_element)
      if elements:
        return [{} for _ in elements]

    for_count:int = self._summary_conf.get(CrawlerName.Field.FOR_COUNT.value) or 1
    # pad a empty entity for each for count
    return [{} for _ in range(for_count)]
    
  def _crawl(self)->STDOut:
    '''
    override the crawl method
    execute the main crawler looply, by the entities or count
    @return {STDOut}
    '''
    self._entities:list[dict] = self._get_entities()

    repeat_count:int = int(self._summary_conf.get('repeat_count') or 1)
    results:list[any] = []
    for _ in range(repeat_count):
      output:STDOut = self._crawl_loop()
      if output.code != 200:
        return output
      
      results = results + output.data
    return STDOut(200,'ok',results)

  def _crawl_loop(self)->STDOut:
    '''
    override the crawl method
    execute the main crawler looply, by the entities or count
    @return {STDOut}
    '''
    if not self._crawler_meta:
      message = f'[{self.NAME}] Failed to crawl - Missing crawler config'
      return STDOut(500,message)

    if not self._entities:
      message = f'[{self.NAME}] Failed to crawl - Missing entities'
      return STDOut(500,message)
    
    try:
      results:list[any] = []
      index = 0
      for entity in self._entities:
        idx_entity = {}
        idx_entity['index'] = index
        index += 1
        idx_entity['no'] = index

        # replace the entity's index placeholders
        entity = HoconReplacer(entity,idx_entity).format()
        # replace the bizdata's index placeholders
        biz = HoconReplacer(self._bizdata,idx_entity).format()
        
        # use the entity to cover the bizdata
        merged_entity = {**biz,**entity,**idx_entity}
        model = Model(self._crawler_meta,merged_entity)
        output:STDOut = self._invoke(model)
        # only save the data field
        if isinstance(output.data,list):
          results.extend(output.data)
        else:
          results.append(output.data)
        self._set_interval()
        
      return STDOut(200,'ok',results)
    except Exception as e:
      message = f'[{self.NAME}] Failed to crawl - {e}'
      return STDOut(500,message)
    


from blues_lib.behavior.Trigger import Trigger

class Scroll(Trigger):

  def _trigger(self)->any:
    kwargs = self._get_kwargs(['loc_or_elem','parent_loc_or_elem'])
    amount_x = self._config.get('amount_x') or 0
    amount_y = self._config.get('amount_y') or 0

    if amount_x or amount_y:
      kwargs = {**kwargs,'amount_x':amount_x,'amount_y':amount_y}
      return self._browser.action.wheel.scroll_from_element_to_offset(**kwargs)
    else:
      return self._browser.action.wheel.scroll_element_to_center(**kwargs)



import sys,os,re

from .NAPS import NAPS

from blues_lib.plan.PublishPlanFactory import PublishPlanFactory     
from blues_lib.model.models.BaiJiaDBModelFactory import BaiJiaDBModelFactory
from blues_lib.loginer.factory.BaiJiaLoginerFactory import BaiJiaLoginerFactory   
from blues_lib.sele.ai.DouBaoImgGen import DouBaoImgGen

class DouBaoImgGenToBaiJiaEvents(NAPS):

  CHANNEL = 'baijia'

  def _get_plan(self):
    return PublishPlanFactory().create_baijia({
      'events':1,
    })
    
  def _get_loginer(self):
    loginer_factory = BaiJiaLoginerFactory()
    return loginer_factory.create_persistent_mac()

  def _get_models(self):
    query_condition = {
      'mode':'latest',
      'material_type':'gallery',
      'count':self._plan.current_total,
    }
    factory = BaiJiaDBModelFactory()
    return factory.create_events(query_condition)

  def spide(self):
    '''
    Crawl a material
    Return:
      {bool}
    '''
    ai = DouBaoImgGen()
    ai.insert()

 





from abc import ABC, abstractmethod
class FuncHandler(ABC):
  def __init__(self,hook_def:dict[str,any],options:dict) -> None:
    '''
    @param {dict} hook_def : hook definition
    @param {dict} options: 可选参数，支持：
        - 其他未来可能扩展的参数（如 context、config 等）
    '''
    super().__init__()
    self._hook_def = hook_def
    self._options = options # 可变的扩展参数
    self._func = eval(self._hook_def.get('value') or 'lambda x:x')

  @abstractmethod
  def execute(self):
    pass



import sys,os,re

from .NAPS import NAPS

from blues_lib.plan.PublishPlanFactory import PublishPlanFactory     
from blues_lib.model.models.BaiJiaDBModelFactory import BaiJiaDBModelFactory
from blues_lib.loginer.factory.BaiJiaLoginerFactory import BaiJiaLoginerFactory   
from blues_lib.schema.reader.doubao.DouBaoSchemaFactory import DouBaoSchemaFactory     
from blues_lib.sele.spider.MaterialSpider import MaterialSpider    

class DouBaoGalleryToBaiJiaEvents(NAPS):

  CHANNEL = 'baijia'

  def _get_plan(self):
    return PublishPlanFactory().create_baijia({
      'events':1,
    })
    
  def _get_loginer(self):
    loginer_factory = BaiJiaLoginerFactory()
    return loginer_factory.create_persistent_mac()

  def _get_models(self):
    query_condition = {
      'mode':'latest',
      'material_type':'gallery',
      'count':self._plan.current_total,
    }
    factory = BaiJiaDBModelFactory()
    return factory.create_events(query_condition)

  def spide(self):
    '''
    Crawl a material
    Return:
      {bool}
    '''
    factory = DouBaoSchemaFactory()
    schemas = factory.create_gallery_list()
    spider = MaterialSpider(schemas,self._plan.current_total,persistent=True)
    return spider.spide() 

 





from bs4 import BeautifulSoup

class BluesSoup():
  
  def __init__(self,driver):
    self.driver = driver
    self.soup = None

  def next(self):
    '''
    @description : get the current page soup
    '''
    html = self.driver.page_source
    self.soup = BeautifulSoup(html,'lxml')

  def find(self,name):
    if not self.soup:
      self.next()
    return self.soup.find(name)
  


from blues_lib.namespace.CommandName import CommandName
from blues_lib.command.NodeCommand import NodeCommand
from blues_lib.type.output.STDOut import STDOut
from blues_lib.cleaner.file.FileCleanerChain import FileCleanerChain   
from blues_lib.cleaner.db.DBCleanerChain import DBCleanerChain

class Cleaner(NodeCommand):

  NAME = CommandName.Standard.CLEANER

  def _invoke(self)->STDOut:

    db_request:dict = {
      'rule':self._summary.get('db_rule'),
      'messages':[],
    }
    file_request:dict = {
      'rule':self._summary.get('file_rule'),
      'messages':[],
    }
    DBCleanerChain().handle(db_request)
    FileCleanerChain().handle(file_request)
    return self._get_output(db_request.get('messages'),file_request.get('messages'))
    
  def _get_output(self,db_messages:list[str],file_messages:list[str])->STDOut:
    return STDOut(200,'ok',db_messages+file_messages)



import sys,os,re
from .ChromeArgs import ChromeArgs

from blues_lib.type.file.File import File

class ChromeExpArgs(ChromeArgs):

  # 传入设置的目录分隔线必须与系统要求一致
  __download_dir = os.path.normpath(File.get_dir_path('browser_download'))
  __exclusion = [
    'enable-automation', # 隐藏“Chrome 正受到自动测试软件控制”提示栏。在自动化测试中，避免被网站检测到自动化工具。
    'ignore-certificate-errors', # 忽略 SSL 证书错误。在测试或爬虫中，访问使用 HTTPS 协议的网站时使用。
  ] 
  __perfs = {
    # 通知设置
    "profile.default_content_settings.popups": 0,       # 禁止弹窗（防止某些站点拦截下载）
    'profile.default_content_setting_values.notification': 2, # 通知权限 0 - Default, 1 - Allow, 2 - Block
    'profile.default_content_setting_values.popups': 0, # 控制网页中通过 window.open() 或 <a target="_blank"> 触发的弹窗窗口  0 - Default, 1 - Allow, 2 - Block
    
    # 界面设置
    'autofill.profile_enabled':False, # 启用自动填充表单, False（禁用）
    
    # 下载配置
    'download.default_directory':__download_dir, # 下载目录, 默认 "C:\\Users\\Downloads"
    'download.prompt_for_download':False, # 是否提示下载路径, False（禁用提示，直接下载到默认目录）
    'download.directory_upgrade':True, # 自动升级下载目录权限
    'download.extensions_to_open':'', # 移除所有直接打开文件类型, such as: "pdf"（直接打开 PDF 文件，不下载）
    'safebrowsing.enabled':False, # 启用安全浏览（屏蔽危险文件）
    "safebrowsing.disable_download_protection": True,   # 允许下载"危险文件"（如 .exe）
    
    # 插件与扩展相关
    'plugins.plugins_disabled':[
      "Chrome PDF Viewer",       # 禁用内置 PDF 预览
      "Adobe Flash Player",      # 禁用 Flash
    ], # 禁用指定插件（需插件 ID）
    'plugins.always_open_pdf_externally':True, # PDF 直接下载（不预览）
    'profile.default_content_setting_values.plugins':1, # 插件全局权限,1（允许）或 2（阻止）
    
    # 资源加载
    'profile.default_content_setting_values.images':0, # 用户级别的图片加载控制,设置为 2 会阻止网页中的 <img> 图片加载, 不阻止 CSS 背景图（如 background-image）、Base64 内联图片、Canvas 绘制的图像。
    'profile.managed_default_content_settings.stylesheets': 0, # 阻止 CSS 加载
    'profile.managed_default_content_settings.javascript': 0, # 阻止 JavaScript 执行
    
  }

  __default_args = {
    'detach':True, # 保持浏览器窗口在脚本结束后不关闭。在调试或需要手动检查页面时使用。
    'excludeSwitches':__exclusion, # 排除项,禁用指定的功能项目
    'useAutomationExtension':False, # 禁用自动化扩展，避免被检测到自动化工具。在需要绕过反爬虫机制或自动化检测的场景中使用。
    'prefs':__perfs, # 设置浏览器的首选项（Preferences）用于配置浏览器的默认行为，如禁用图片、设置下载目录等。
  }
    
  def get(self):
    '''
    Get the default and input experimental args
    @returns {dict}
    '''
    args = {**self.get_from_input()}
    return args
  
  def get_from_input(self):
    args = {}
    if not self._input_args:
      return args

    # 直接使用入参 
    for key,value in self._input_args.items():
      args[key] = value
    return args



from blues_lib.behavior.Trigger import Trigger

class Switch(Trigger):

  def _trigger(self)->bool:
    '''
    switch to the specified window handle
    @returns {bool}
    '''
    try:
      if not self._browser:
        return True

      handle:str = self._config.get('handle')
      if handle == 'default' or not handle:
        self._browser.interactor.window.switch_to_default()
      elif handle=='latest':
        self._browser.interactor.window.switch_to_latest()
      elif handle=='prev':
        self._browser.interactor.window.switch_to_prev()
      elif handle=='next':
        self._browser.interactor.window.switch_to_next()
      else:
        self._browser.interactor.window.switch_to(handle)
      return True
    except Exception as e: 
      return False


from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model
from blues_lib.namespace.CrawlerName import CrawlerName
from blues_lib.crawler.NestCrawler import NestCrawler

class DepthCrawler(NestCrawler):

  NAME = CrawlerName.Engine.DEPTH

  def _before_crawled(self):
    super()._before_crawled()
    # crawler
    self._children:list[dict] = self._meta.get(CrawlerName.Field.CHILDREN.value,[])
    # url
    self._pages:list[dict] = self._get_pages()
    # calculate
    self._max_depth:int = len(self._children)
    self._rows:list[dict] = []
    
  def _get_pages(self)->list[dict]:
    pages:list[dict] = self._summary_conf.get(CrawlerName.Field.PAGES.value,[])
    if pages:
      return [page for page in pages if page.get('url')]
    
    urls:list[str] = self._summary_conf.get(CrawlerName.Field.URLS.value,[])
    url:str = self._summary_conf.get(CrawlerName.Field.URL.value)
    urls = urls if urls else ([url] if url else [])
    if not urls:
      return urls

    pages:list[dict] = []
    for url in urls:
      if not url:
        continue
      pages.append({
        "url":url,
        "bizdata":{}, # must be a dict
      })

    return pages
    
  def _crawl(self)->STDOut:
    '''
    override the crawl method
    @return {STDOut}
    '''
    if not self._children:
      message = f'[{self.NAME}] Failed to crawl - Missing children config'
      return STDOut(500,message)
    
    if not self._pages:
      message = f'[{self.NAME}] Failed to crawl - Missing pages config'
      return STDOut(500,message)

    # loop out of the dps, make sure Deep First Search
    depth = 1
    for page in self._pages:
      model = self._get_urls_replaced_model(page,depth)
      self._dfs(model,depth)

      if self._count != -1 and len(self._rows) >= self._count:
        break
      
    if self._rows:
      return STDOut(200,'success',self._rows)
    else:
      message = f'[{self.NAME}] Failed to crawl - No available entities found'
      return STDOut(500,message)

  def _dfs(self,model:Model,depth:int):

    # crawl : loop crawler will merge the entity to the output.data
    output:STDOut = self._crawl_by_meta(model,CrawlerName.Engine.LOOP.value)
    if output.code!=200 or not output.data:
      return

    if depth == self._max_depth:
      # add the entity to the rows
      self._rows.extend(output.data)
      return 
    
    next_depth = depth+1
    # loop for Deep First Search
    for entity in output.data:
      model = self._get_entities_replaced_model(entity,next_depth)
      self._dfs(model,next_depth)
      # must set be break, avoid to crawl useless mat
      if self._count != -1 and len(self._rows) >= self._count:
        break
      
  def _get_entities_replaced_model(self,entity:dict,depth:int)->Model:
    bizdata = {
      **self._bizdata, 
      CrawlerName.Field.ENTITIES.value:[entity],
    } 

    meta = self._children[depth-1]
    return Model(meta,bizdata)

  def _get_urls_replaced_model(self,page:dict,depth:int)->Model:
    url:str = page.get('url')
    page_bizdata:dict = page.get('bizdata')
    bizdata = {
      **self._bizdata,
      **page_bizdata, # page bizdata cover the main bizdata
      CrawlerName.Field.URLS.value:[url], 
    } 

    meta = self._children[depth-1]
    return Model(meta,bizdata)




from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model
from blues_lib.namespace.CrawlerName import CrawlerName
from blues_lib.crawler.base.BaseCrawler import BaseCrawler

class UrlCrawler(BaseCrawler):

  NAME = CrawlerName.Engine.URL
    
  def _crawl(self)->STDOut:
    '''
    override the crawl method
    @return {STDOut}
    '''
    if not self._crawler_meta:
      message = f'[{self.NAME}] Failed to crawl - Missing crawler config'
      return STDOut(500,message)

    # must pass the meta and bizdata, some behavior need to calculate the model
    model = Model(self._crawler_meta,self._bizdata)

    return self._invoke(model)



from blues_lib.material.file.MatFile import MatFile
from blues_lib.cleaner.CleanerHandler import CleanerHandler
from blues_lib.util.BluesFiler import BluesFiler

class LogHandler(CleanerHandler):

  kind = 'handler'

  def resolve(self,request)->int:
    '''
    Args:
      {dict} request : 
        - {int} validity_days : by default is 100
    Returns {int} count
    '''
    validity_days = request.get('rule',{}).get('validity_days',30)
    root = MatFile.get_log_root()
    count = BluesFiler.removedirs(root,validity_days) or 0
    request.get('messages').append(f'Deleted {count} log files')
    return count


import sys,os,re

from .NAPS import NAPS

from blues_lib.plan.PublishPlanFactory import PublishPlanFactory     
from blues_lib.model.models.TouTiaoDBModelFactory import TouTiaoDBModelFactory
from blues_lib.loginer.factory.TouTiaoLoginerFactory import TouTiaoLoginerFactory   

class TouTiaoEvents(NAPS):

  CHANNEL = 'toutiao'

  def _get_plan(self):
    return PublishPlanFactory().create_toutiao({
      'events':1,
    })
    
  def _get_loginer(self):
    loginer_factory = TouTiaoLoginerFactory()
    return loginer_factory.create_persistent_mac()

  def _get_models(self):
    query_condition = {
      'mode':'latest',
      'material_type':'article',
      'count':self._plan.current_total,
    }
    factory = TouTiaoDBModelFactory()
    return factory.create_events(query_condition)



from blues_lib.namespace.CommandName import CommandName
from blues_lib.command.SQLCommand import SQLCommand
from blues_lib.dao.sql.TableMutator import TableMutator
from blues_lib.type.output.STDOut import STDOut

class Inserter(SQLCommand):

  NAME = CommandName.SQL.INSERTER

  def _invoke(self)->STDOut:
    table = self._summary.get('table')
    entities = self._summary.get('entities')

    filter_rule:dict = self._summary.copy()
    filtered_entities = self._filter_entities(entities,filter_rule)

    mutator = TableMutator(table)
    return mutator.insert(filtered_entities)



from pymysql.converters import escape_string

class BluesSQLConvertor():

  @classmethod
  def get_in_sql(cls,condition):
    '''
    @description : convert list or tuple to in sql
    @param {dict} condition  
      {'field':'name','values':['a','b','c']}
    @returns {str} : in sql
    '''
    sql = ''
    field = condition.get('field','')
    values = condition.get('values')

    if not values:
      return ''

    for value in values:
      sql += '"%s",' % value 
    return ' %s in (%s) ' % (field,sql[:-1])

  @classmethod
  def get_order_sql(cls,orders):
    '''
    @description : return order by sql
    @param {dict[]} : orders
      [{'field':'id','sort':'asc'},{'field':'name','sort':'desc'}] 
    @returns {str} : order by sql
    '''
    sql = ''
    if not orders:
      return sql

    order_list = [orders] if isinstance(orders,dict) else orders 
    for order in order_list:
      field = order.get('field')
      sort = order.get('sort')

      if not field:
        continue

      sql += '%s %s,' % (field,sort)

    if not sql:
      return ''

    return ' order by %s ' % sql[:-1] 

  @classmethod
  def get_update_template_sql(cls,fields,conditions):
    '''
    @description : convert dict to update sql
    @param {list|tuple} fields : the fields will be udpated
      ['name','age']
    @param {dict[]} conditions : one or more conditions
      [
        {'operator':'and','field':'name','comparator':'=','value':'blues','value_type':'str'},
        {'operator':'and','field':'name','comparator':'=','value':'blues'}
      ] 
    @returns {str} : update sql
    '''
    sql = ''
    condition_sql = cls.get_condition_sql(conditions)
    for field in fields:
      sql += field+'=%s,' # don't wrapper by ""
    return ' set %s %s ' % (sql[:-1],condition_sql)

  @classmethod
  def get_update_sql(cls,entity,conditions):
    '''
    @description : convert dict to update sql
    @param {dict} entity : the entity dict's key is the real table field
      {'name':'blues','age':18}
    @param {dict[]} conditions : one or more conditions
      [
        {'operator':'and','field':'name','comparator':'=','value':'blues'},
        {'operator':'and','field':'name','comparator':'=','value':'blues'}
      ] 
    @returns {str} : update sql
    '''
    sql = ''
    condition_sql = cls.get_condition_sql(conditions)
    for field in entity:
      value = entity[field]
      # 处理字符串类型：需要转义并加引号
      if isinstance(value, str):
        escaped_value = escape_string(value)
        sql += f"{field}='{escaped_value}',"
      # 处理数字、布尔值等非字符串类型：直接拼接，不加引号
      else:
        sql += f"{field}={value},"
    return ' set %s %s ' % (sql[:-1],condition_sql)

  @classmethod
  def get_insert_template_sql(cls,fields):
    '''
    @description : convert dict to insert sql
    @param {str[]} fields : the fields will be insert
      ['name','age']
    @returns {str} : insert template sql
    '''
    key_sql = ''
    value_sql = ''
    
    for field in fields:
      key_sql += '%s,' % field
      value_sql += '%s,'

    return ' (%s) values (%s) ' % (key_sql[:-1],value_sql[:-1])

  @classmethod
  def get_insert_sql(cls,entities):
    '''
    @description : convert dict to insert sql
    @param {dict|dict[]} entities : the entity dict's key is the real table field
      [{'name':'blues','age':18}]
    @returns {str} : insert sql
    '''
    entity_list = [entities] if isinstance(entities,dict) else entities
    key_sql = ''
    value_sql = ''
    
    first_entity = entity_list[0]
    for key in first_entity:
      key_sql += '%s,' % key

    for entity in entity_list:    
      unit_value_sql = ''
      for key in entity:
        # must use escape_string to deal Quote collision problem
        # 处理可能的非字符串类型，仅对字符串进行转义
        if isinstance(entity[key], str):
          escape_value = escape_string(entity[key]) if entity[key] else ''
        else:
          # 非字符串类型（如int、float等）直接使用原始值，空值处理为''
          escape_value = entity[key] if entity[key] else ''

        unit_value_sql += '"%s",' % escape_value
      value_sql += '(%s),' %  unit_value_sql[:-1]

    return ' (%s) values %s ' % (key_sql[:-1],value_sql[:-1])

  @classmethod
  def get_condition_sql(cls,conditions):
    '''
    @description : get sql conditons
    @param {dict[]} conditions : one or more conditions
      [
        {'operator':'and','field':'name','comparator':'=','value':'blues'},
        {'operator':'and','field':'name','comparator':'=','value':'blues'}
      ] 
    @returns {str} : conditon sql
    '''
    sql = ''
    if not conditions:
      return sql 

    avail_condition_count = 0
    condition_list = [conditions] if isinstance(conditions,dict) else conditions 
    for condition in condition_list:
      field = condition.get('field')
      comparator = condition.get('comparator','=')
      operator = condition.get('operator','and')
      value_type = condition.get('value_type')

      if comparator == 'in':
        original_value = condition.get('value')
        if not original_value:
          continue

        value = cls.get_in_sql(field,original_value)
      else:
        value = condition.get('value','')

      if not field:
        continue

      if avail_condition_count==0:
        operator = ''

      avail_condition_count+=1
      if comparator=='in':
        sql += ' %s ( %s ) ' % (operator,value)
      else:
        if value_type=='function':
          sql += ' %s (%s %s %s) ' % (operator,field,comparator,value)
        else:
          sql += ' %s (%s %s "%s") ' % (operator,field,comparator,value)

    return ' where %s ' % sql

  @classmethod
  def get_limit_sql(cls,pagination):
    '''
    @description : get limit sql
    @param {dict} pagination : page info
      {'no':1,'size':10}
    @returns {str} : limit sql
    '''
    if not pagination:
      return ''

    no = int(pagination.get('no',1))
    size = int(pagination.get('size',50))
    from_index = (no-1)*size
    return ' limit %s,%s ' % (from_index,size)
    



import os,time,logging
from abc import ABC,abstractmethod
from openai import OpenAI
from blues_lib.type.output.STDOut import STDOut

class Completion(ABC):
  _BASE_URL = "https://api.deepseek.com"
  _API_KEY_ENV_NAME = 'DEEPSEEK_API_KEY'

  _MODEL_NAME = 'deepseek-chat'
  _TEMPERATURE = 1
  _STREAM = False
  _RESPONSE_FORMAT={
    'type': 'text'
  }
  
  def __init__(self,request_body:dict|None=None, api_key: str = '') -> None:
    self._request_body = request_body or {}
    self._api_key = api_key
    self._logger = logging.getLogger('airflow.task')

  def ask(self,messages:list[dict[str,str]])->STDOut:
    kwargs = self._get_kwargs(messages)
    try:
      client = self._get_client()
      self._console(messages)

      request_start = time.time()
      response = client.chat.completions.create(**kwargs)

      request_end = time.time()
      self._logger.info(f"API处理时间: {request_end - request_start:.2f}秒")

      stdout:STDOut = self._get_content(response)
      self._logger.info(f"llm output: {stdout}")
      return stdout
    except Exception as e:
      return self._get_error(e)
  
  def _get_kwargs(self,messages:list[dict[str,str]]):
    kwargs = {
      'stream':self._STREAM,
      'temperature':self._TEMPERATURE,
      'model':self._MODEL_NAME,
      'response_format':self._RESPONSE_FORMAT,
    }
    kwargs.update(self._request_body)
    # messages动态输入
    kwargs['messages'] = messages
    return kwargs
  
  def _console(self,messages:list[dict[str,str]]):
    system_prompt = messages[0]['content']
    user_prompt_len = len(messages[1]['content'])
    self._logger.info(f'llm prompt : {system_prompt.strip()[:50]}...({user_prompt_len} chars)')
    
  def _get_client(self):
    return OpenAI(
      # 优先使用客户端传入，再次使用环境变量
      api_key= self._api_key or os.environ.get(self._API_KEY_ENV_NAME,''),
      base_url=self._BASE_URL,
      timeout=120,  # 长文本处理需要更长时间
      max_retries=3,  # 添加重试机制
    )

  def _get_error(self,error)->STDOut:
    common_err:dict = {
      "code":"common_error",
      "type":"internal_error",
      "message":str(error)
    }

    std_err:dict|None = None 
    # 异常为OpenAIError对象，需要解析为字典使用
    if hasattr(error,'response'):
      response:dict = error.response.json()
      # 避免内部结构变动异常
      std_err = response.get('error')

    std_err = std_err or common_err
    return STDOut(
      500,
      f"{std_err.get('code')} - {std_err.get('type')} - {std_err.get('message')}",
    )

  @abstractmethod
  def _get_content(self,response)->STDOut:
    pass


from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.by import By

class XPathFinder():

  '''
  Locating the elements based on the provided locator values.
  '''

  def __init__(self,driver):
    self.__driver = driver
  
  # === part 1:  get element by text === #
  def find(self, target_xpath:str, parent_element:WebElement=None) -> WebElement:

    """
    根据xpath查找元素，可指定父级选择器
    
    :param target_xpath: 要查找的xpath
    :param parent_element: 父级元素，默认为None表示从根元素开始查找
    :return: 找到的WebElement对象
    """
    # 拼接XPath，先定位父级再在其内部查找包含指定文本的元素
    return self._find_by_xpath(target_xpath,parent_element)

  def find_all(self, target_xpath:str, parent_element:WebElement=None) -> list[WebElement]:

    """
    根据xpath查找所有元素，可指定父级选择器
    
    :param target_xpath: 要查找的xpath
    :param parent_element: 父级元素，默认为None表示从根元素开始查找

    :return: 找到的WebElement对象列表
    """
    return self._find_all_by_xpath(target_xpath,parent_element)


  # === appendix:  private methods === #
  def _find_by_xpath(self,xpath,parent_element:WebElement=None):
    try:
      if parent_element:
        return parent_element.find_element(By.XPATH,xpath)
      else:
        return self.__driver.find_element(By.XPATH,xpath)
    except:
      return None


  def _find_all_by_xpath(self,xpath,parent_element:WebElement=None):
    try:
      if parent_element:
        return parent_element.find_elements(By.XPATH,xpath)
      else:
        return self.__driver.find_elements(By.XPATH,xpath)
    except:
      return None



class JSEvent():

  def click(self,selector,parent_selector=''):
    sel = selector
    if parent_selector:
      sel = '%s %s' % (parent_selector,selector)
    script = 'document.querySelector("%s").click();' % sel
    self.execute(script)

  



from typing import Type
from blues_lib.hook.Hook import Hook
from .CommandProcFactory import CommandProcFactory
from .CommandFuncHandler import CommandFuncHandler
from blues_lib.hook.HookExcept import HookExcept
from blues_lib.type.output.STDOut import STDOut

class CommandHook(Hook):
  
  def _get_proc_factory(self)->Type[CommandProcFactory]:
    return CommandProcFactory
  
  def _get_func_handler(self)->Type[CommandFuncHandler]:
    return CommandFuncHandler
  
  def execute(self)->None:
    if not self._hook_defs:
      return

    HookExcept.validate(self._hook_defs)

    ti = self._options.get('ti')
    for hook_def in self._hook_defs:
      # if one hook block or skip, then break
      if ti.xcom_pull(key='should_skip') or ti.xcom_pull(key='should_block'):
        break
      self._run(hook_def)

  @classmethod
  def block(cls,ti:any)->None:
    if ti.xcom_pull(key='should_block'):
      stdout:STDOut = STDOut(500,f'The task {ti.task_id} is blocked',False)
      for key,value in stdout.to_dict().items():
        ti.xcom_push(key,value)
      raise Exception(f'The task {ti.task_id} is blocked')
    
  @classmethod
  def skip(cls,ti:any)->STDOut|None:
    if ti.xcom_pull(key='should_skip'):
      stdout:STDOut = STDOut(200,f'The task {ti.task_id} is skipped',True)
      for key,value in stdout.to_dict().items():
        ti.xcom_push(key,value)
      return stdout
    return None
    



from selenium.webdriver.common.by import By

class Locator:

  bys:dict = {
    'css_selector':By.CSS_SELECTOR,
    'xpath':By.XPATH,
    'id':By.ID,
    'name':By.NAME,
    'class_name':By.CLASS_NAME,
    'tag_name':By.TAG_NAME,
    'link_text':By.LINK_TEXT,
    'partial_link_text':By.PARTIAL_LINK_TEXT,
  }  

  @classmethod
  def get(cls,loc_or_elem,by:str='')->tuple[str,str]|None:
    '''
    Split the by and value from the loc_or_elem,such as:
    'css_selector::id=123' -> (By.CSS_SELECTOR,'id=123')
    'xpath:://div' -> (By.XPATH,'//div')
    if has no found the by_key,return By.CSS_SELECTOR
    
    @param {str} loc_or_elem - The target CS_WE,such as 'css_selector::id=123'
    @param {str} by - The by type,such as 'css selector','xpath', it's the valid value ,not the dict key
    '''
    if not loc_or_elem:
      return None

    by_value:str = ''
    locator_value:str = ''
    default_by_value:str = by or By.CSS_SELECTOR
    # only split the first '::',such as 'css_selector::id=123' -> ['css_selector','id=123']
    locator_list:list[str] = loc_or_elem.split('::',1)
    
    if len(locator_list) == 1:
      by_value = default_by_value
      locator_value = loc_or_elem
    else:
      by_key = locator_list[0] 
      by_value = cls.bys.get(by_key) or default_by_value
      locator_value = locator_list[1]
      
    return (by_value,locator_value)




from blues_lib.model.Model import Model
from blues_lib.behavior.Bean import Bean
from blues_lib.behavior.BhvExecutor import BhvExecutor

class Richtext(Bean):

  def _set(self)->any:
    # [{'type':'text','value':'xxx'},{'type':'image','value':'c:/xx.png'}]
    paras:list[dict] = self._config.get('value')
    if not paras or not isinstance(paras,list):
      return

    frame_loc_or_elem:str = self._config.get('frame_loc_or_elem')
    if frame_loc_or_elem:
      self._switch(frame_loc_or_elem,'framein')
      
    self._setup()

    prev_para = {} 
    for para in paras:
      self._set_para(para,prev_para)
      prev_para = para

    if frame_loc_or_elem:
      self._switch(frame_loc_or_elem,'frameout')
      
  def _setup(self):
    conf:dict = self._config.get('setup',{})
    bhv_chain:list[dict] = conf.get('bhv_chain')
    if not conf or not bhv_chain:
      return

    frame_loc_or_elem:str = conf.get('frame_loc_or_elem')
    if frame_loc_or_elem:
      self._switch(frame_loc_or_elem,'framein')
      
    model = Model(bhv_chain)
    BhvExecutor(model,self._browser).execute()

    if frame_loc_or_elem:
      self._switch(frame_loc_or_elem,'frameout')
      

  def _set_para(self,para:dict,prev_para:dict)->None:
    type = para.get('type')
    prev_type = prev_para.get('type')
    value = para.get('value')
    conf:dict = self._config.get(type,{})
    
    meta:dict = self._meta.get(type,{})
    bhv_chain:list[dict] = meta.get('bhv_chain')
    if not value or not bhv_chain:
      return
    
    frame_loc_or_elem:str = conf.get('frame_loc_or_elem')
    if frame_loc_or_elem:
      self._switch(frame_loc_or_elem,'framein')
    
    # if the prev para is image, then add a empty line before the text
    if type=='text' and prev_type=='image':
      value = ['',value]

    bizdata = {
      **self._bizdata,
      "value":value
    }
    model = Model(bhv_chain, bizdata)
    BhvExecutor(model,self._browser).execute()

    if frame_loc_or_elem:
      self._switch(frame_loc_or_elem,'frameout')
  
  def _switch(self,frame_loc_or_elem:str,kind:str)->None:
    conf = {
      '_kind':kind,
      'loc_or_elem':frame_loc_or_elem
    }
    model = Model(conf)
    BhvExecutor(model,self._browser).execute()



class JSJQuery():

  def is_jquery_available(self):
    return self.is_var_available('jQuery')
    
  def load_jquery(self):
    cdn = 'https://libs.baidu.com/jquery/2.0.0/jquery.min.js'
    self.load_script(cdn)



from abc import abstractmethod
from typing import Any
from blues_lib.hocon.replacer.Replacer import Replacer

class BaseReplacer(Replacer):
  ### macro statement replacer 

  IS_VARIABLES_REQUIRED = False
  
  def replace(self)->dict:
    """递归替换所有占位符"""
    if not self._template:
      return self._template
    
    if self.IS_VARIABLES_REQUIRED and not self._variables:
      return self._template

    return self._replace_node(self._template)
  
  def _replace_node(self, node:Any)->Any:
    """递归遍历并替换占位符"""
    if isinstance(node, str):
      return self._replace_node_var(node)
    elif isinstance(node, dict):
      # 创建新字典，不修改原对象
      return {key: self._replace_node(value) for key, value in node.items()}
    elif isinstance(node, list):
      # 处理列表
      return [self._replace_node(item) for item in node]
    else:
      # 其他类型直接返回
      return node
  
  @classmethod
  @abstractmethod
  def _replace_node_var(cls, value:str)->str:
    """替换单个字符串中的占位符（要求完全匹配）"""
    pass


import random,time
from selenium.webdriver.common.keys import Keys

from blues_lib.sele.waiter.Querier import Querier  

class Input():

  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver,5)

  def write(self,loc_or_elem,value,parent_loc_or_elem=None,timeout=5):
    '''
    Clear and write text into the text controller
    Parameter:
      loc_or_elem {str | WebElement} : the input element's css selecotr or web element
      texts {list<str>} : one or more text string
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return None

    self.clear(web_element,None)

    self.append(web_element,value,None)

  def append(self,loc_or_elem,value,parent_loc_or_elem=None,timeout=5):
    '''
    Append text into the text controller
    Parameter:
      loc_or_elem {str | WebElement} : the input element's css selecotr or web element
      texts {list<str>} : one or more text string
    '''
    texts = value if type(value)==list else [value]

    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return None

    web_element.send_keys(*texts)

  def write_para(self,loc_or_elem,value,LF_count=1,parent_loc_or_elem=None,timeout=5):
    '''
    Write lines with line break
    Parameter:
      loc_or_elem {str | WebElement} : the input element's css selecotr or web element
      parent_loc_or_elem {str | WebElement} : the input element parent's css selecotr or web element
      texts {list<str>} : texts
      LF_count {int} : line break count in every para
      input_by_para {bool} : input txt para by para
    '''
    texts = value if type(value)==list else [value]
    paras = self.__get_paras(texts,LF_count)

    self.clear(loc_or_elem,parent_loc_or_elem,timeout)
    for para in paras:
      self.append(loc_or_elem,para,parent_loc_or_elem,timeout)

  def append_para(self,loc_or_elem,value,LF_count=1,parent_loc_or_elem=None,timeout=5):
    texts = value if type(value)==list else [value]
    paras = self.__get_paras(texts,LF_count)
    for para in paras:
      self.append(loc_or_elem,para,parent_loc_or_elem,timeout)

  def __get_paras(self,texts,LF_count):
    break_texts = []
    idx = 0
    max_idx = len(texts)-1
    for text in texts:
      break_texts.append(text)
      if idx<=max_idx:
        for i in range(LF_count):
          break_texts.append(Keys.ENTER)
      idx+=1
    return break_texts

  def write_discontinuous(self,loc_or_elem,value,parent_loc_or_elem=None,timeout=5,interval:list[int|float]|None=None):

    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return None

    self.clear(web_element,None)
    self.append_discontinuous(web_element,value,None,timeout,interval)
  
  def append_discontinuous(self,loc_or_elem,value,parent_loc_or_elem=None,timeout=5,interval:list[int|float]|None=None):

    '''
    Input chars non-uniform speed
    '''
    texts = value if type(value)==list else [value]
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)

    for text in texts:
      # input char by char
      for char in text:
        self.__input_discontinuous(web_element,char,interval)

  def __input_discontinuous(self,web_element,char,interval:list[int|float]|None=None):
    '''
    input the text one char by one char intermittently
    using a random interval
    '''
    if interval:
      time.sleep(random.uniform(interval[0],interval[1]))
    web_element.send_keys(char)

  def clear(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    """
    清空元素内容，支持表单输入框和所有可编辑元素（含div、span等）
    """
    element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    tag_name = element.tag_name.lower()
    
    # 处理表单输入元素
    if tag_name in ['input', 'textarea']:
      element.clear()
    else:
      # 处理所有可编辑元素（不限制标签类型）
      contenteditable = element.get_attribute('contenteditable')
      # contenteditable可能为"true"或""（空字符串表示启用）
      is_editable = contenteditable in (True, 'true', '')
      if is_editable:
        self.__driver.execute_script("arguments[0].innerHTML = '';", element)


from blues_lib.hook.HookProc import HookProc

class BehaviorProc(HookProc):
  
  def __init__(self,proc_def:dict,options:dict) -> None:
    super().__init__()

    self._proc_conf = proc_def
    # options's value will be rewrite
    self._options = options



from urllib.parse import urlparse
from blues_lib.type.chain.AllMatchHandler import AllMatchHandler
from blues_lib.material.file.MatFile import MatFile
from blues_lib.util.BluesAlgorithm import BluesAlgorithm
from blues_lib.util.BluesURL import BluesURL

class Downloader(AllMatchHandler):
  # download the images to the local

  def resolve(self)->None:
    entity = self._request.get('entity')

    if thumb_url := entity.get('mat_thumb'):
      entity['mat_thumb'] = self._download_thumb(entity,thumb_url)

    if paras:=entity.get('mat_paras'):
      entity['mat_paras'] = self._download_body_images(entity,paras)
    
  def _download_thumb(self,entity:dict,thumb_url)->str:
    return self._download(entity,thumb_url,'thumb')

  def _download_body_images(self,entity:dict,paras:list)->list[dict]:
    image_idx = 1
    convert_paras = []
    for para in paras:
      if para['type'] == 'image':
        if local_image := self._download(entity,para['value'],f'body-{image_idx}'):
          image_idx += 1
          para['value'] = local_image
          convert_paras.append({**para})
      else:
        convert_paras.append({**para})
    return convert_paras
    
  def _download(self,entity:dict,image_url:str,file_name_suffix:str='')->str:
    if not image_url:
      self._logger.warning(f'[{self.__class__.__name__}] Skip a empty url')
      return ''

    if not BluesURL.is_http_url(image_url):
      self._logger.warning(f'[{self.__class__.__name__}] Skip a not http url - {image_url}')
      return image_url

    # 有限使用页面url中域名作为命名空间
    mat_url = entity.get('mat_url')
    mat_site = entity.get('mat_site') or BluesURL.get_main_domain(mat_url)
    mat_id = entity.get('mat_id') or BluesAlgorithm.md5(entity['mat_url'])
    file_name = f"{mat_id}-{file_name_suffix}"

    stdout:STDOut = MatFile.get_download_image(mat_site,file_name,image_url)
    if stdout.code!=200:
      self._logger.error(f'[{self.__class__.__name__}] {stdout.message} - {stdout.data}')
      return ''
    return stdout.data



from blues_lib.material.MatHandler import MatHandler
from blues_lib.type.output.STDOut import STDOut

class Filter(MatHandler):

  def _calculate(self)->STDOut:
    valid_entities = []
    invalid_entities = []

    for item in self._entities:
      output:STDOut = self._insert(item)
      if output.code !=200:
        item['mat_stat'] = 'invalid'
        item['mat_remark'] = output.message
        invalid_entities.append(item)
      else:
        valid_entities.append(item)

    message:str = f"valid {len(valid_entities)} ; invalid {len(invalid_entities)}" 
    return STDOut(200,message,valid_entities,invalid_entities)




import os,logging,datetime

from blues_lib.util.BluesFiler import BluesFiler 
from blues_lib.material.file.MatFile import MatFile

class LoggerSetter(): 

  _LOG_LEVELS = {
    'debug':logging.DEBUG,
    'info':logging.INFO, # will print all system info 
    'warning':logging.WARNING,
    'error':logging.ERROR,
  }
  _WIDTH = 100 
  _NAME = 'airflow.task' # same to the logger name in the DAG

  def __init__(self,config={}):
    '''
    @description : set log config
    @param {str} config.directory : Directory for storing logs
    @param {str} config.name : log's topic
    @param {int} config.retention_days : Maximum number of days for storing logs
    @param {str} config.level : The lowest level of log output
      - enum: debug info warning error
    '''
    self._config = self._get_config(config)


  def _get_config(self,config:dict):
    dft = {
      'name':self._NAME,
      'file_name':os.environ.get('OS_LOG_NAME',self._NAME),
      'directory':MatFile.get_today_log_root(),
      'level':'info',
      'retention_days':7,
    }
    return {**dft,**config} if config else dft
    
  def _set_dir(self):
    # make sure the dir exist
    BluesFiler.makedirs(self._config['directory'])
    # clear log history
    BluesFiler.removefiles(self._config['directory'],self._config['retention_days'])

  def set(self):
    self._set_dir()

    log_level  = self._LOG_LEVELS.get(self._config['level'],logging.INFO)
    logging.basicConfig(level=log_level)
    logger = logging.getLogger(self._config['name']) # will be show in the formatter as the `name`

    formatter = self._get_airflow_formatter()

    # output to the logs
    log_file = self._get_log_file()
    file_logger = logging.FileHandler(log_file,'a','utf-8')
    file_logger.setFormatter(formatter)
    logger.addHandler(file_logger)
    
    # output to the console
    #console_logger = logging.StreamHandler()
    #console_logger.setFormatter(formatter)
    #logger.addHandler(console_logger)

  def _get_airflow_formatter(self):
    formatter = '[%(asctime)s] %(filename)s %(lineno)s %(levelname)s - %(message)s'
    return logging.Formatter(formatter,datefmt='%Y-%m-%d %H:%M:%S')
      
  def _get_formatter(self):
    split_line = ''.join(['-' for x in range(self._WIDTH)])
    formatter = split_line+'\n'+'%(levelname)s (%(name)s) %(asctime)s:\n%(message)s'
    return logging.Formatter(formatter)
      
  def _get_log_file(self):
    today = datetime.datetime.now().strftime("%Y-%m-%d")
    log_name = f"{self._config['file_name']}_{today}.log"
    return os.path.join(self._config['directory'],log_name)

  @property
  def config(self):
    return self._config

  @property
  def file(self):
   return self._get_log_file() 

  def print_file(self):
    title = self.file
    char = '-'
    padding = (self._WIDTH - len(title) - 2) // 2
    left = char * padding
    right = char * (self._WIDTH - padding - len(title) - 2)
    print(f'\n{left} {title} {right}')




import time
from selenium.webdriver.remote.webelement import WebElement
from blues_lib.behavior.Trigger import Trigger
from blues_lib.types.common import LocOrElems

class Click(Trigger):

  def _trigger(self)->int:
    loc_or_elems:LocOrElems = self._config.get('loc_or_elems')
    if loc_or_elems:
      return self._multi_click(loc_or_elems)
    else:
      return self._click() 
  
  def _multi_click(self,loc_or_elems:LocOrElems)->int:
    index = 0
    if isinstance(loc_or_elems,str):
      items = self._get_eles(loc_or_elems)
    else:
      items = loc_or_elems
    for loc_or_elem in items:
      self._config['loc_or_elem'] = loc_or_elem
      self._click()
      if index<len(items)-1:
        self._wait_interval(index)
      index += 1
    return index
  
  def _get_eles(self,locator:str)->list[WebElement]:
    kwargs:dict = self._get_kwargs(['parent_loc_or_elem','timeout'])
    kwargs['loc_or_elem'] = locator
    return self._browser.waiter.querier.query_all(**kwargs)

  def _click(self):
    kwargs:dict = self._get_kwargs(['loc_or_elem','parent_loc_or_elem','timeout'])
    if self._is_clickable():
      self._scroll()
      time.sleep(0.2) # 有时状态判断成功，但实际还需要等待一些
      return self._browser.action.mouse.click(**kwargs)

  def _is_clickable(self)->bool:
    loc_or_elem = self._config.get('loc_or_elem')
    if isinstance(loc_or_elem,WebElement):
      return True
    else:
      return self._to_be_clickable()



from .ChromeArgs import ChromeArgs

# Chrome DevTools Protocal 浏览器启动后动态设置,必须在get打开页面之前设置
class ChromeCDPArgs(ChromeArgs):

  __default_args = {
    'Network.setUserAgentOverride':{
      # 设置用户代理
      'userAgent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'
    },
    # 移除navigator上的webdriver标识
    'Page.addScriptToEvaluateOnNewDocument':{
      'source': "Object.defineProperty(navigator, 'webdriver', {get: () => undefined});"
    },
  }
    
  def get(self):
    '''
    Get the default and input experimental args
    @returns {dict} : removed duplicate args
    '''
    return {**self.__default_args,**self.get_from_input()}
  
  def get_from_input(self):
    '''
    Convert the pass standard args settings to the real args
      - replace the input value to the placehoder
    '''
    args = {}
    if not self._input_args:
      return args
    
    for key,value in self._input_args.items():
      args[key] = value # value maybe 0
    return args



from blues_lib.behavior.Trigger import Trigger

class Hover(Trigger):

  def _trigger(self):
    kwargs = self._get_kwargs(['loc_or_elem','parent_loc_or_elem','timeout'])
    if self._to_be_visible():
      return self._browser.action.mouse.move_in(**kwargs)


from enum import Enum

class NSEnum(Enum):

  # overriede __str__
  def __str__(self):
    return self.value
  
  @classmethod
  def from_value(cls, value: str):
    """通过字符串值获取对应的枚举成员"""
    try:
      # 直接使用 Enum 的内置机制
      return cls(value)  
    except ValueError:
      return None




from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions 
from selenium.common.exceptions import NoSuchElementException 
from selenium.common.exceptions import ElementNotInteractableException 
from .deco.ExpecterDeco import ExpecterDeco

class Expecter():
  '''
  Define custom expected conditions
  '''

  def __init__(self,driver):
    self.__driver = driver
    self.__timeout = 10
    self.__poll_frequency = 0.5
    self.__ignored_exceptions = [NoSuchElementException,ElementNotInteractableException]
    self.__timeout_message = 'timeout'

  #-- module 9: wait by self-defined func --#
  @ExpecterDeco('until')
  def until(self,func,timeout=5,poll_frequency=None,ignored_exceptions=None,timeout_message=None):
    # this func has a para is the driver
    wait_timeout = timeout if timeout else self.__timeout
    wait_poll_frequency = poll_frequency if poll_frequency else self.__poll_frequency
    wait_ignored_exceptions = ignored_exceptions if ignored_exceptions else self.__ignored_exceptions
    wait_timeout_message = timeout_message if timeout_message else self.__timeout_message
    
    # can return any type result
    return WebDriverWait(self.__driver,wait_timeout,wait_poll_frequency,wait_ignored_exceptions).until(func,wait_timeout_message)

  def __demo_func(driver):
    '''
    The first parameter (driver) must be accepted
    This func's return value will be the until's return value
    '''
    return driver.current_url



import jsonschema
from blues_lib.schema.SchemaReader import SchemaReader

class SchemaValidator:
  """
  Class for validating metadata against JSON Schema templates
  """
  @classmethod
  def validate(cls, instance:any, schema:dict)->tuple[bool,str]:
    stat,message = SchemaValidator.validate_schema(schema)
    if not stat:
      return False, message

    try:
      jsonschema.validate(instance, schema)
      return True, 'validated'
    except jsonschema.exceptions.ValidationError as e:
      message:str = cls._get_message(e)
      return False, message
    except Exception as e:
      return False, str(e)

  @classmethod
  def validate_as_metadta(cls, instance:any, )->tuple[bool,str]:
    '''
    validate the metadata with a schema template
    - every metadata must has a 'tpl_id' field
    '''
    tpl_path:str = instance.get('tpl_id')
    if not tpl_path:
      return False, "metadata must has a 'tpl_id' field"
    return cls.validate_with_template(instance, tpl_path)

  @classmethod
  def validate_with_template(cls, instance:any, tpl_path:str)->tuple[bool,str]:
    '''
    Validate instance against JSON Schema template
    Args:
      instance: Instance to validate
      tpl_path: Path to JSON Schema template, such as 
        - 'dag' in root dir
        - 'metadata.llm' in sub dir
        - 'dag.json' with extension
    Returns:
      tuple[bool,str]: Validation result and message
    '''
    schema = SchemaReader.read(tpl_path)
    if not schema:
      return False, f"Schema template {tpl_path} not found"
    return cls.validate(instance, schema)
    
  @classmethod
  def _get_message(cls, validation_error):
    """
    Get validation error message
    Args:
        validation_error: jsonschema ValidationError object
    Returns:
        str: Formatted error message
    """
    # Get error path
    path = '/'.join(map(str, validation_error.path))
    return f"{validation_error.message} - at {path}"
    
  @classmethod
  def validate_schema(cls, schema)->tuple[bool,str]:
    """
    Validate if a schema is a valid JSON Schema
    Args:
      schema: Schema to validate
    Returns:
      bool: True if valid
    """
    try:
      jsonschema.validators.validator_for(schema).check_schema(schema)
      return True,''
    except Exception as e:
      return False, str(e)
    


from functools import wraps
from blues_lib.type.output.STDOut import STDOut
from blues_lib.deco.LogDeco import LogDeco

class BehaviorHookLog(LogDeco):

  def __init__(self,title:str='hook',max_chars:int=100):
    super().__init__(title,max_chars)

  def __call__(self,func):
    @wraps(func) 
    def wrapper(instance,*args,**kwargs):
      prefix:str = self._get_prefix(instance)
      stdout:STDOut = func(instance,*args,**kwargs)
      self._log(prefix,stdout)
      return stdout
    return wrapper

  def _get_prefix(self,instance)->str:
    module_name:str = instance.__class__.__module__  # 模块路径（如 my_project.tasks.loop）
    return f'{module_name} [{self._title}]'


from selenium.webdriver.remote.webelement import WebElement
from blues_lib.sele.waiter.EC import EC   
from blues_lib.sele.waiter.deco.QuerierDeco import QuerierDeco
from blues_lib.sele.element.Finder import Finder   

# 提供元素选择功能
class Querier():

  def __init__(self,driver,timeout=8):
    self.__driver = driver
    self.__ec = EC(driver) 
    self.__finder = Finder(driver) 
    self.__timeout = timeout or 5

  def setTimeout(self,timeout=5):
    '''
    Adjust the timeout in runtime
    '''
    self.__timeout = timeout

  @QuerierDeco('query')
  def query(self,loc_or_elem,parent_loc_or_elem=None,timeout=5)->WebElement|None:
    '''
    Wait and get the element from document or parent element
    Parameter:
      loc_or_elem {str|WebElement} : the target element's css selector or WebElement
      parent_loc_or_elem {str|WebElement} : the parent element's css selector or WebElement
      timeout {int} : Maximum waiting time (s)
    Returns:
      {WebElement|None} 
    '''
    if not loc_or_elem:
      return None
    if not parent_loc_or_elem:
      return self.__query(loc_or_elem,timeout)

    parent:WebElement|None = self.__query(parent_loc_or_elem,timeout) 
    if parent:
      return self.__finder.find(loc_or_elem,parent)
    else:
      return self.__query(loc_or_elem,timeout)

  @QuerierDeco('query_all')
  def query_all(self,loc_or_elem,parent_loc_or_elem=None,timeout=5)->list[WebElement]|None:
    '''
    Wait and get elements from document or parent element
    Parameter:
      loc_or_elem {str|WebElement} : the target element's css selector or WebElement
      parent_loc_or_elem {str|WebElement} : the parent element's css selector or WebElement
      timeout {int} : Maximum waiting time (s)
    Returns:
      {list<WebElement>} 
    '''
    if not loc_or_elem:
      return None
    
    if not parent_loc_or_elem:
      return self.__query_all(loc_or_elem,timeout)

    parent:WebElement|None = self.__query(parent_loc_or_elem,timeout)
    if parent:
      return self.__finder.find_all(loc_or_elem,parent)
    else:
      return self.__query_all(loc_or_elem,timeout)

  def __query(self,loc_or_elem,timeout=5,parent_loc_or_elem=None)->WebElement|None:
    '''
    Wait and Get the target WebElement
    Parameter:
      loc_or_elem {str|WebElement} : the target element's css selector or WebElement
      timeout {int} : Maximum waiting time (s)
    Returns:
      {WebElement|None} 
    '''
    if isinstance(loc_or_elem,WebElement):
      return loc_or_elem
    
    wait_time = timeout or self.__timeout
    return self.__ec.to_be_presence(loc_or_elem,wait_time,parent_loc_or_elem)

  def __query_all(self,loc_or_elem,timeout=5,parent_loc_or_elem=None)->list[WebElement]|None:
    '''
    Wait and Get the target WebElements
    Parameter:
      loc_or_elem {str|WebElement} : css selector or web element
      timeout {int} : Maximum waiting time (s)
    Returns:
      {list<WebElement>} 
    '''
    if isinstance(loc_or_elem,WebElement):
      return [loc_or_elem]

    wait_time = timeout or self.__timeout
    return self.__ec.all_to_be_presence(loc_or_elem,wait_time,parent_loc_or_elem)

   



from blues_lib.sele.browser.chrome.ChromeFactory import ChromeFactory   
from blues_lib.crawler.CrawlerFactory import CrawlerFactory
from blues_lib.sele.browser.Browser import Browser
from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model

class Invoker():
  
  @classmethod
  def invoke(cls,model:Model,context:dict|None=None)->STDOut:
    browser = cls.get_browser(model.config,context)
    crawler = CrawlerFactory(model,browser).create()
    return crawler.execute()
  
  @classmethod
  def get_browser(cls,config:dict,context:dict|None=None)->Browser:
    #if context and context.get('browser'):
    #  return context.get('browser')

    browser_conf:dict = config.get('browser') or {}
    driver_config =  browser_conf.get('config') or {} # optional
    driver_options = browser_conf.get('options') or {} # optional
    browser = ChromeFactory(**driver_options).create(driver_config)
    if not browser:
      raise Exception(f'[{self.NAME}] fail to create the browser')
    return browser


from copy import deepcopy
from blues_lib.command.FlowCommand import FlowCommand
from blues_lib.type.output.STDOut import STDOut

class LoopFlowCommand(FlowCommand):

  NAME = None

  def _run(self)->STDOut:

    loop:dict = self._task_def.get('loop') or {}
    entities:list[dict] = loop.get('entities') or []
    count:int = int(loop.get('count') or -1)
    
    if not entities:
      raise Exception(f'{self.NAME} must have loop entities')

    # one by one
    items:list[dict] = []
    for entity in entities:
      # must pass a list
      if not entity:
        continue

      stdout:STDOut = self._run_one([entity])
      sub_items:list[dict] = stdout.data or []
      items.extend(sub_items)
      if count > 0 and len(items) >= count:
        break
    items = items[:count] if count > 0 else items
    return STDOut(200,'flow loop success',items)

  def _run_one(self,start_data:list[dict])->STDOut:
    dag_def:dict = self._get_def_with_start()
    dag_biz:list[dict] = self._get_biz_with_start(start_data)
    return self._run_flow(dag_def,dag_biz)
  
  def _get_biz_with_start(self,start_data:list[dict])->list[dict]:
    start_biz = {"data":start_data}
    dag_biz:list[dict] = self._bizdata.get('tasks')
    return [start_biz]+dag_biz

  def _get_def_with_start(self)->dict:
    dag_def:dict = deepcopy(self._dag_def)
    start_def:dict = self._get_start_task()
    tasks:list[dict] = dag_def.get('tasks') or []

    dag_def['tasks'] = [start_def] + tasks
    return dag_def
    
  def _get_start_task(self)->dict:
    return {
      "task_id":"start",
      "command":"command.standard.dummy",
      "meta":{
        "summary":{
          "code":200,
          "message":"dummy start",
          "data":"{{data}}",
          "detail":None
        }
      },
    }



from blues_lib.type.output.STDOut import STDOut
from blues_lib.behavior.Behavior import Behavior
from blues_lib.deco.BehaviorSTDOutLog import BehaviorSTDOutLog

class Bean(Behavior):

  @BehaviorSTDOutLog()
  def _invoke(self)->STDOut:
    value = None
    try:
      if self._action()=='setter':
        value = self._set()
      else:
        value = self._get()
        value = self._after_invoked(value)
      return STDOut(200,'ok',value)
    except Exception as e:
      return STDOut(500,str(e),value)

  def _get(self)->any:
    # optional
    pass

  def _set(self)->any:
    # optional
    pass

  def _action(self)->str:
    action = 'getter'
    if 'value' in self._config:
      action = 'setter'
    return action
  
  def _get_value_entity(self)->dict|None:

    key:str|None = self._config.get('key')
    value:any = self._config.get('value')

    if key:
      return {
        key:value
      }

    if isinstance(value,dict):
      return value
    
    return None
      


import sys,os,re

from blues_lib.util.BluesDateTime import BluesDateTime
from blues_lib.behavior.Bean import Bean
from blues_lib.dao.login.LoginQuerier import LoginQuerier
from blues_lib.model.Model import Model
from blues_lib.type.output.STDOut import STDOut

class AuthCodeInput(Bean):

  def _set(self)->bool:
    code = self._config['value']
    if not code or code == '__sentinel__':
      code = self._get()

    if not code:
      self._logger.error(f'[{self.__class__.__name__}] Failed to fetch the auth code')
      return False

    stdout = self._input(code)
    
    if stdout.code==200:
      self._logger.info(f'[{self.__class__.__name__}] Managed to fetch and input the auth code - {code}')
      return True
    else:
      self._logger.error(f'[{self.__class__.__name__}] Failed to input the auth code - {code} - {stdout.message}')
      return False

  def _input(self,code)->STDOut:
    from blues_lib.behavior.BhvExecutor import BhvExecutor

    resetable = self._config.get('resetable',False)
    input_conf = {
      '_kind':'reset_char' if resetable else 'write_char',
      'loc_or_elem':self._config.get('loc_or_elem'),
      'value':code,
    }

    model = Model(input_conf)
    executor = BhvExecutor(model,self._browser)
    return executor.execute()

  def _get(self)->str:
    '''
    Wait the code upload and continue
    '''
    step = 10
    code_expire = self._config.get('code_expire',300)
    domain = self._config.get('domain')
    ts = self._config.get('timestamp')
    steps =  list(range(0,code_expire,step)) 
    i = 0
    for s in steps:
      i+=1
      
      if code := self._get_code(domain,ts):
        return str(code) # must be a str

      BluesDateTime.count_down({
        'duration':step,
        'title':f'[{self.__class__.__name__}] Wait for the auth code {i*step}/{code_expire} - {domain} {ts}'
      })
    return ''

  def _get_code(self,domain,ts)->str:
    conditions = [
      {'field':'login_ts','comparator':'=','value':ts},
      {'field':'login_site','comparator':'=','value':domain},
    ]
    querier = LoginQuerier()
    stdout:STDOut = querier.get('*',conditions)
    data = stdout.data
    return data[0]['login_sms_code'] if data else ''




from abc import ABC, abstractmethod

class Replacer(ABC):
  
  def __init__(self,template:dict, variables:dict|None=None):
    self._template = template
    self._variables = variables

  @abstractmethod
  def replace(self)->dict:
    pass
  


from blues_lib.behavior.write.Write import Write

class WriteText(Write):

  _set_keys = ['loc_or_elem','value','parent_loc_or_elem','timeout']

  def _set(self)->any:
    kwargs = self._get_kwargs(self._set_keys)
    clearable:bool = self._config.get('clearable',False)
    # must wait the element visible, or can't use the send_keys method
    if self._to_be_visible():
      if clearable:
        return self._browser.element.input.write(**kwargs)
      else:
        return self._browser.element.input.append(**kwargs)


from blues_lib.namespace.CommandName import CommandName
from blues_lib.command.NodeCommand import NodeCommand
from blues_lib.type.output.STDOut import STDOut
from blues_lib.material.validator.Validator import Validator as ValidatorHandler

class Validator(NodeCommand):
  NAME = CommandName.Material.VALIDATOR

  def _invoke(self)->STDOut:
    '''
    验证字段是否合法，长度要比llm要求更宽一些
    '''
    rule:dict = self._summary.get('rule',{})
    entities:list[dict] = self._summary.get('entities',[])
    if not entities:
      raise ValueError(f'{self.NAME} : no input entities')
    
    request = {
      'rule':rule,
      'entities':entities, # must be a list
    } 
    handler = ValidatorHandler(request)
    return handler.handle()


from blues_lib.hook.command.CommandProc import CommandProc
from blues_lib.deco.CommandHookLog import CommandHookLog
from blues_lib.type.output.STDOut import STDOut

class MatProc(CommandProc):
  
  @CommandHookLog()
  def execute(self)->None:
    rule:dict = self._proc_conf.get('rule',{})
    entities:list[dict] = self._stdout.data or []

    request = {
      'rule':rule,
      'entities':entities, # must be a list
    } 
    stdout:STDOut = self._calculate(request)

    self._stdout.message = stdout.message
    self._stdout.data = stdout.data
    self._stdout.trash = stdout.trash

  def _calculate(self,request:dict)->STDOut:
    raise NotImplementedError



from blues_lib.util.ResourceReader import ResourceReader
from blues_lib.util.DotFileReader import DotFileReader

class SchemaReader:
  
  PACKAGE = 'blues_lib.schema'

  @classmethod
  def read(cls, dot_path:str) -> dict:
    extension:str = 'conf'
    slash_path:str = DotFileReader.get_slash_path(dot_path,extension)
    return ResourceReader.read_hocon(cls.PACKAGE, slash_path) 



import sys,os,re
from .deco.SwitchDeco import SwitchDeco

from blues_lib.sele.waiter.Querier import Querier

class Frame():
 
  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver)
  
  @SwitchDeco('Frame','switch_to',1)
  def switch_to(self,loc_or_elem):
    '''
    Switch the driver's focus to the specific frame
    Parameter:
      loc_or_elem {str|WebElement} : the frame's css selector or webelement
    Returns:
      {None}
    '''
    web_element = self.__querier.query(loc_or_elem)
    if not web_element:
      return False
    try:
      self.__driver.switch_to.frame(web_element)
      return True
    except:
      return False
      
  @SwitchDeco('Frame','switch_to_default')
  def switch_to_default(self):
    '''
    Switch to the default window that the driver opened
    Returns:
      {None}
    '''
    try:
      self.__driver.switch_to.default_content()
      return True
    except:
      return False

  @SwitchDeco('Frame','switch_to_parent')
  def switch_to_parent(self):
    '''
    Switch to the current frame's parent window, mybe not the default window
    Returns:
      {None}
    '''
    try:
      self.__driver.switch_to.parent_frame()
      return True
    except:
      return False

  @SwitchDeco('Frame','execute',1)
  def execute(self,loc_or_elem,func):
    '''
    Switch to a frame, execute a function and back to the default window
    Parameter:
      loc_or_elem {str|WebElement} : the frame's css selector or webelement
      func {function} : ececute in the iframe window
    Returns:
      {any} : the func's return value
    '''
    self.switch_to(loc_or_elem)
    value = func()
    self.switch_to_default()
    return value
      



from pyhocon import ConfigFactory
from pyhocon.config_tree import ConfigTree
from pyhocon.tool import HOCONConverter

class HoconReader:

  @classmethod
  def read(cls,file_path:str)->ConfigTree:
    '''
    @description: read conf file as ConfigTree
    @param {str} file path such as 'c:/blues-lib-py/tests/mock/command/llm-loop-urls/def.conf'
    @returns {ConfigTree}
    '''
    try:
      return ConfigFactory.parse_file(file_path)
    except Exception as e:
      print(f"read hocon error: {e}")
      return ConfigTree()

  @classmethod
  def read_as_dict(cls,file_path:str)->dict:
    '''
    @description: read conf file as dict
    @param {str} file path such as 'c:/blues-lib-py/tests/mock/command/llm-loop-urls/def.conf'
    @returns {dict}
    '''
    config:ConfigTree = cls.read(file_path)
    return config.as_plain_ordered_dict()

  @classmethod
  def read_as_hocon(cls,file_path:str)->str:
    '''
    @description: read conf file as hocon string
    @param {str} file path such as 'c:/blues-lib-py/tests/mock/command/llm-loop-urls/def.conf'
    @returns {str}
    '''
    config:ConfigTree = cls.read(file_path)
    return HOCONConverter.convert(config, 'hocon')

  @classmethod
  def read_as_json(cls,file_path:str)->str:
    '''
    @description: read conf file as json string
    @param {str} file path such as 'c:/blues-lib-py/tests/mock/command/llm-loop-urls/def.conf'
    @returns {str}
    '''
    config:ConfigTree = cls.read(file_path)
    return HOCONConverter.convert(config, 'json')
  
  @classmethod
  def read_as_yaml(cls,file_path:str)->str:
    '''
    @description: read conf file as yaml string
    @param {str} file path such as 'c:/blues-lib-py/tests/mock/command/llm-loop-urls/def.conf'
    @returns {str}
    '''
    config:ConfigTree = cls.read(file_path)
    return HOCONConverter.convert(config, 'yaml')
  
  @classmethod
  def read_as_properties(cls,file_path:str)->str:
    '''
    @description: read conf file as properties string
    @param {str} file path such as 'c:/blues-lib-py/tests/mock/command/llm-loop-urls/def.conf'
    @returns {str}
    '''
    config:ConfigTree = cls.read(file_path)
    return HOCONConverter.convert(config, 'properties')



import time
from blues_lib.behavior.Trigger import Trigger

class ToBeClickable(Trigger):

  def _trigger(self)->bool:
    '''
    if the current url is equal to the expected url
    @returns {bool}
    '''
    kwargs = self._get_kwargs(['loc_or_elem','timeout','parent_loc_or_elem'])
    stat:bool = self._browser.waiter.ec.to_be_clickable(**kwargs)

    if stat and (post_gurard_time:= self._config.get('post_guard_time',1)):
      time.sleep(post_gurard_time)
    return stat



from blues_lib.behavior.Trigger import Trigger

class EleAbsents(Trigger):

  def _trigger(self)->bool:
    '''
    check if the element is present in the page
    @returns {bool}
    '''
    loc_or_elem:str = self._config.get('loc_or_elem')
    wait_time:int = self._config.get('wait_time',3)
    return self._browser.waiter.ec.to_be_stale(loc_or_elem,wait_time)




class NestedDataUpdater:
  """嵌套数据写入器，通过路径字符串合并数据"""
  
  @staticmethod
  def update(data,value):
    if not data or not value:
      return 
    if isinstance(data,list) and isinstance(value,list):
      data.extend(value)
    elif isinstance(data,dict) and isinstance(value,dict):
      data.update(value)

  @staticmethod
  def write_by_path(data: any, path: str, value: any) -> bool:
    """
    通过路径写入数据，类型不匹配时返回False（不写入）
    
    参数:
        data: 原始数据（可以是字典、列表或其他可嵌套的数据结构）
        path: 目标路径字符串，支持字典属性和列表下标，例如 'summary.query'、'0.summary.query.1.answer'
        value: 要写入的值
    
    返回:
        是否成功写入
    """
    if not path:
      NestedDataUpdater.update(data,value)
      return True  # if hava no path, assign the value to the root data
    
    parts = path.split('.')
    parent_parts = parts[:-1]
    last_part = parts[-1]
    
    # 找到父节点
    current: any = data
    for part in parent_parts:
      if isinstance(current, list):
        try:
          index = int(part)
          if 0 <= index < len(current):
            current = current[index]
          else:
            return False
        except ValueError:
          return False
      elif isinstance(current, dict):
        if part in current:
          current = current[part]
        else:
          return False
      else:
        return False  # 中间节点不是可访问类型
    
    # 写入最后一个节点
    if isinstance(current, dict):
      # 字典允许任何字符串键
      NestedDataUpdater.update(current[last_part],value)
      return True
    elif isinstance(current, list):
      try:
        index = int(last_part)
        if 0 <= index < len(current):
          NestedDataUpdater.update(current[index],value)
          return True
        else:
          return False  # 列表索引越界
      except ValueError:
        return False  # 列表索引必须是整数
    else:
      return False  # 父节点不是字典或列表，无法写入


from blues_lib.dao.login.LoginMutator import LoginMutator
from blues_lib.cleaner.CleanerHandler import CleanerHandler
from blues_lib.type.output.SQLSTDOut import SQLSTDOut

class LoginLogHandler(CleanerHandler):

  kind = 'handler'
  mutator = LoginMutator()

  def resolve(self,request)->int:
    '''
    Args:
      {dict} request: 
        - {int} validity_days : by default is 100
    Returns {int} count
    '''
    validity_days = request.get('rule',{}).get('validity_days',100)
    conditions = [
      {
        'field':'login_created_time',
        'comparator':'<=',
        'value':'DATE_SUB(CURRENT_DATE, INTERVAL %s DAY)' % validity_days,
        'value_type':'function',
      }
    ]
    output:SQLSTDOut = self.mutator.delete(conditions)
    count = output.count or 0
    request.get('messages').append(f'Deleted {count} login log rows')
    return count


import sys,os,re

from blues_lib.type.output.STDOut import STDOut
from blues_lib.type.chain.Handler import Handler

class FirstMatchHandler(Handler):

  def handle(self):
    stdout = self.resolve()
    if stdout.code==200:
      return stdout

    if self._next_handler:
      return self._next_handler.handle()

    return STDOut(500,'Failed to handle the request',None)



from abc import ABC,abstractmethod
from blues_lib.command.NodeCommand import NodeCommand
from blues_lib.model.Model import Model
from blues_lib.type.output.STDOut import STDOut
from blues_lib.hook.command.CommandHook import CommandHook
from blues_lib.deco.CommandSTDOutLog import CommandSTDOutLog

class LoopCommand(NodeCommand,ABC):

  NAME = None

  def _setup(self)->bool: 
    super()._setup()
    self._loop:dict = self._task_def.get('loop') or {}
    self._entities:list[dict] = self._loop.get('entities') or []
    self._map:dict = self._loop.get('map') or {}
    self._count:int = int(self._loop.get('count') or -1)
    self._round:int = 0
  
  def _invoke(self)->STDOut:
    if self._entities:
      return self._run()
    else:
      return self._run_once(self._model)
  
  def _run(self)->STDOut:
    items:list[dict] = []
    for entity in self._entities:
      model = self._get_loop_model(entity)
      stdout:STDOut = self._run_once(model,entity)
      if not stdout.data:
        continue

      items.extend(stdout.data)
      if self._count > 0 and len(items) >= self._count:
        break
    items = items[:self._count] if self._count > 0 else items
    return STDOut(200,'loop success',items)

  @CommandSTDOutLog('loop',max_chars=100)
  def _run_once(self,model:Model,entity:dict|None=None)->STDOut:
    self._round += 1
    
    self._before_each_invoked()
    output:STDOut = self._run_once_cal(model)
    self._merge(output,entity)
    
    self._after_each_invoked(output)
    return output
  
  @abstractmethod
  def _run_once_cal(self,model:Model)->STDOut:
    pass

  def _get_loop_model(self,entity:dict)->Model:
    # don't update the original bizdata
    bizdata:dict = self._model.bizdata.copy() 
    meta:dict = self._model.meta.copy()

    if self._map:    
      for entity_key,bizdata_key in self._map.items():
        bizdata[bizdata_key] = entity.get(entity_key)
    else:
      # bizdata and entity has the same field names
      for key,value in entity.items():
        bizdata[key] = value
    
    return Model(meta,bizdata)
  
  def _get_loop_items(self,items:list[dict],entity:dict)->list[dict]:
    filter_entity = {k:v for k,v in entity.items() if k not in self._map.keys()}
    merged_items:list[dict] = []
    for item in items:
      # llm输出优先
      merged = {**filter_entity,**item}
      merged_items.append(merged)
    return merged_items
  
  def _merge(self,output:STDOut,entity:dict|None=None)->None:
    code:int = output.code
    data:any = output.data
    if code == 200 and data:
      items:list[dict] = data if isinstance(data,list) else [data]
      merged_items = self._get_loop_items(items,entity) if entity else items
      output.data = merged_items
    else:
      output.data = []

  def _before_each_invoked(self)->None:
    hook_defs:list[dict] = self._task_def.get('before_each_invoked')
    options:dict = {'ti':self._ti}
    CommandHook(hook_defs,options).execute()

  def _after_each_invoked(self,output:STDOut)->None:
    # update the items directly 
    hook_defs:list[dict] = self._task_def.get('after_each_invoked')
    options:dict = {'ti':self._ti,'stdout':output}
    CommandHook(hook_defs,options).execute()



from blues_lib.behavior.Trigger import Trigger

class ToBeSelected(Trigger):

  def _trigger(self)->bool:
    '''
    if the current url is equal to the expected url
    @returns {bool}
    '''
    kwargs = self._get_kwargs(['loc_or_elem','timeout'])
    return self._browser.waiter.ec.to_be_selected(**kwargs)



from blues_lib.behavior.Trigger import Trigger
from blues_lib.util.BluesMailer import BluesMailer  

class Email(Trigger):

  def _trigger(self)->bool:
    payload = self._get_payload()
    mailer = BluesMailer.get_instance()
    stdout = mailer.send(payload)
    return stdout.code == 200

  def _get_payload(self)->dict:
    payload_conf:dict = self._config.get('value')
    subject = payload_conf.get('subject')
    paras = payload_conf.get('paras')
    addressee = payload_conf.get('addressee') or ['langcai10@dingtalk.com']
    addressee_name = payload_conf.get('addressee_name') or 'BluesLiu'

    return {
      'subject':subject,
      'paras':paras,
      'addressee':addressee, # send to multi addressee
      'addressee_name':addressee_name,
    }



import sys,os,re
from typing import List

from blues_lib.behavior.Bean import Bean

class BrCookie(Bean):

  def _get(self)->List[dict]:
    # read cookie from the browser
    return self._browser.interactor.cookie.get()

  def _set(self)->bool:
    # get cookie from the local file and set to the browser
    cookies = self._config['value']
    if not cookies or cookies == '__sentinel__':
      cookies = self._browser.read_cookies()
 
    if not cookies:
      self._logger.warning(f'[{self.__class__.__name__}] Failed to read cookies from the local file')
      return False

    self._logger.info(f'[{self.__class__.__name__}] Managed to set cookies to the browser')
    self._browser.interactor.cookie.replace(cookies)
    return True



class JSAlert():

  # == module 1: html or text == #
  def alert(self,text='accept'):
    '''
    Popup the alert dialog
    Parameter:
      text {str} : the alert's text 
    Returns:
      {None} : alert return undefined in js
    '''
    script = '''
      return alert(`%s`);
    ''' % (text)
    return self.execute(script)
  

  def confirm(self,text='accept or dismiss'):
    '''
    Returns:
      {bool} : accept - true; dismiss - false
    '''
    script = '''
      return confirm(`%s`);
    ''' % (text)
    return self.execute(script)
  
  def prompt(self,text='input and confirm'):
    '''
    Returns:
      {str|None} : accept - the input value; dismiss - None
    '''
    script = '''
      return prompt(`%s`);
    ''' % (text)
    return self.execute(script)
  



from blues_lib.behavior.Bean import Bean

class PageSource(Bean):

  def _get(self)->str:
    self._scroll_to_bottom()
    return self._browser.interactor.document.get_page_source()
  
  def _scroll_to_bottom(self):
    # 必须先滚动到底部，确保滚动懒加载资源更新，例如图片，只滚动4次
    scroll_conf=self._config.get('scroll') or {}
    step = scroll_conf.get('step') or 1000
    interval = scroll_conf.get('interval') or 1000
    # 必须至少滚2次/2秒，因为之前名没有设置其他等待或检查
    max_step = scroll_conf.get('max_step') or 2
    self._browser.script.javascript.lazy_scroll_bottom(step,interval,max_step)


import sys,os,re
from typing import Any

from blues_lib.behavior.Bean import Bean

class Select(Bean):

  def _set(self)->Any:
    kwargs = self._get_kwargs(['loc_or_elem','value','parent_loc_or_elem','timeout'])
    if self._to_be_presence():
      return self._browser.element.select.select_by_value_or_text(**kwargs)



from blues_lib.llm.deepseek.JsonChat import JsonChat    
from blues_lib.llm.deepseek.ChatMessages import ChatMessages
from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model

class Invoker():

  @classmethod
  def invoke(cls,model:Model)->STDOut:
    
    prompt:dict = model.config.get('prompt') or {}
    request:dict = model.config.get('request') or {}
    messages = ChatMessages(prompt).create()
    body:dict|None = request.get("body")
    return JsonChat(body).ask(messages)


import sys,os,re,time

from blues_lib.util.BluesURL import BluesURL   
from blues_lib.util.BluesFiler import BluesFiler    

class File():

  LIB_ROOT_PATH = os.environ.get('BLUES_LIB_ROOT_PATH') or BluesFiler.get_lib_path()

  @classmethod
  def get_dir_path(cls,subdir=None):
    '''
    Get the blues lib's standard dir path
    Parameter:
      subdir {None|str|list} : the sub dirs
        None: return the root dir
        str: return the subdir
        list: return the multi subdir
    '''
    if not subdir:
      dir_path = cls.LIB_ROOT_PATH 
    else:
      dirs = subdir if type(subdir)==list else [subdir]
      # support multi level dirs
      dir_path = os.path.join(cls.LIB_ROOT_PATH,*dirs)

    # create dir
    BluesFiler.makedirs(dir_path)
    return dir_path 

  @classmethod
  def get_file_path(cls,subdir,filename):
    '''
    Get the file absolute path base the blues lib root dir
    Support add subdirs
    Parameter:
      subdir {str,list} : one or multi level dirs
      filename {str} : the filename
    '''
    dir_path = cls.get_dir_path(subdir)
    return os.path.join(dir_path,filename)

  @classmethod
  def get_domain_file_path(cls,url,subdir,extension='txt'):
    domain = BluesURL.get_main_domain(url)
    filename = '%s.%s' % (domain,extension)
    return cls.get_file_path(subdir,filename)

  @classmethod
  def get_file_name(cls,prefix='',name='',suffix='',extension=''):
    file_name = name if name else int(time.time()*1000)
    if prefix:
      file_name='%s-%s' % (prefix,file_name)
    if suffix:
      file_name='%s-%s' % (file_name,suffix)
    if extension:
      file_name='%s.%s' % (file_name,extension)
    return file_name 




import logging
from abc import ABC,abstractmethod

from blues_lib.type.output.STDOut import STDOut

class Handler(ABC):

  def __init__(self,request):
    self._next_handler = None
    self._request = request
    self._logger = logging.getLogger('airflow.task')
  
  def set_next(self,handler):
    self._next_handler = handler
    return handler

  @abstractmethod
  def handle(self)->STDOut:
    pass

  @abstractmethod
  def resolve(self)->STDOut:
    pass


class ProcFactory():
  
  _PROC_CLASSES = None

  @classmethod
  def create(cls,hook_def:dict,options:dict):
    class_name:str = hook_def.get('value') or ''
    proc_def:dict = hook_def.get('proc') or {}
    if not class_name:
      return None

    if proc_class := cls._PROC_CLASSES.get(class_name):
      return proc_class(proc_def,options)
    
    return None


import sys,os,re

from blues_lib.util.BluesType import BluesType 
from blues_lib.type.file.File import File   

class Window():
 
  def __init__(self,driver):
    self.__driver = driver

  # == module 1: set window size == #  
  def maximize(self):
    self.__driver.maximize_window()

  def minimize(self):
    self.__driver.minimize_window()

  def fullscreen(self):
    '''
    Maximize and hide the top tool bar
    '''
    self.__driver.fullscreen_window()

  # == module 2: window screenshot == #  
  def screenshot(self,file_path='')->str:
    '''
    Take screenshot and save as image
    Parameter:
      file_path {str} : local download file path
    Returns:
      {str} : the download file path
    '''
    default_path = self.__get_default_file()

    shot_path = file_path if file_path else default_path
    stat = self.__driver.save_screenshot(shot_path)

    return shot_path if stat else ''

  def __get_default_file(self,prefix='screenshot'):
    dl_dir = File.get_dir_path('screenshot') # 已经包含尾部目录分割线
    filename = File.get_file_name(prefix=prefix,extension='png')
    return os.path.join(dl_dir,filename)

  # == module 3: set/get window size == #  
  def get_size(self):
    '''
    Get the window size
    Returns:
      {dict} : the size dict, like:
        - {'width': 1755, 'height': 946}
    '''
    return self.__driver.get_window_size()

  def set_size(self,width,height):
    '''
    Set the window size
    Paraeter:
      width {int} : window's width
      height {int} : window's height
    Returns:
      {None}
    '''
    self.__driver.set_window_size(width,height)

  # == module 4: set/get window position == #  
  def get_position(self):
    '''
    Get the window's position
    Returns:
      {dict} : the position dict,like:
        - {'x': 99, 'y': 49}
    '''
    return self.__driver.get_window_position()
     
  def set_position(self,x,y):
    '''
    Get the window's position
    Parameter:
      {int} x : the distance between the window's left border to the screen's left border
      {int} y : the distance between the window's top border to the screen's top border
    '''
    self.__driver.set_window_position(x,y)

  # == module 5: get handle info == #  
  def get_handle(self):
    '''
    Get the current window id
    Returns:
      {str} ： handle id, like:
        - '3D31BF6D96E5671253E70BCF33DC7F39'
    '''
    return self.__driver.current_window_handle

  def get_handles(self):
    '''
    Get all window's handles
    Returns:
      {list<str>} ： handles list, like:
        - ['3D31BF6D96E5671253E70BCF33DC7F39']
    '''
    return self.__driver.window_handles

  def get_latest_handle(self):
    '''
    Get the latest opened window
    '''
    return self.get_handles()[-1]

  # == module 6: open new window and switch to it  automactially (selenium v4 api) == #  
  def new_window(self,url=''):
    self.__driver.switch_to.new_window('window')
    if url:
      self.__driver.get(url)
      
  def new_tab(self,url=''):
    '''
    Open a new tab / new window and switch to it automatically
    Parameter:
      {str} url : the url will be opened by the new window
    '''
    self.__driver.switch_to.new_window('tab')
    if url:
      self.__driver.get(url)
      
  # == module 7: toggle window == #  
  def switch_to(self,handle_id):
    '''
    Switch to the specified window
    Parameter:
      handle_id {string} 
    Returns:
      {None}
    '''
    self.__driver.switch_to.window(handle_id)

  def switch_to_latest(self):
    '''
    @description : switch window to the latest opened tab
    '''
    latest_handle = self.get_latest_handle()
    current_handle = self.get_handle()
    if latest_handle != current_handle:
      self.switch_to(latest_handle)
  
  def switch_to_default(self):
    '''
    Click a href and open a tab automatically, wouldn't switch the handle
    But the chrome's view toggle to the new tab
    Need switch back manually
    '''
    handles = self.get_handles()
    self.switch_to(handles[0])

  def switch_to_prev(self):
    '''
    @description : switch window to the prev tab
    '''
    handles = self.get_handles()
    current_handle = self.get_handle()
    current_handle_index = BluesType.last_index(handles,current_handle)
    if current_handle_index>0:
      prev_handle = handles[current_handle_index-1]
      self.switch_to(prev_handle)

  def switch_to_next(self):
    '''
    @description : switch window to the prev tab
    '''
    handles = self.get_handles()
    current_handle_index = BluesType.last_index(handles,self.get_handle())
    if current_handle_index<len(handles)-1:
      next_handle = handles[current_handle_index+1]
      self.switch_to(next_handle)



import logging
from selenium import webdriver
import undetected_chromedriver as udc_driver
from selenium.webdriver.chrome.service import Service
from pathlib import Path
from blues_lib.sele.browser.Browser import Browser
from blues_lib.sele.browser.chrome.OptionsCreator import OptionsCreator
from blues_lib.sele.browser.driver.CFTInstaller import CFTInstaller

class Chrome(Browser):
  
  _DEFAULT_PAGE_LOAD_TIMEOUT = 60 # 页面加载超时时间
  _DEFAULT_SCRIPT_TIMEOUT = 60 # 脚本执行超时时间

  def __init__(self,
      std_args=None, # {dict} standard args
      exp_args=None, # {dict} experimentalargs
      cdp_args=None, # {dict} chrome devtools protocal args
      sel_args=None, # {dict} selenium args
      ext_args=None, # {dict} extension args
      executable_path='', # {str} driver.exe path: 'env' - using the env; 'xxx' - the local path; None - using the driver manager
      remote_hub='', # {str} 远程 Selenium hub 服务地址 - http://localhost:4445/wd/hub 
      undetected=True, # {bool} 是否使用 undetected_chromedriver
      page_load_timeout=None, # {int} 页面加载超时时间
      script_timeout=None, # {int} 脚本执行超时时间
    ):
    self._std_args = std_args or {}
    self._exp_args = exp_args or {}
    self._cdp_args = cdp_args or {}
    self._sel_args = sel_args or {}
    self._ext_args = ext_args or {}
    self._executable_path = executable_path
    self._remote_hub = remote_hub
    self._undetected = undetected
    self._page_load_timeout = page_load_timeout or self._DEFAULT_PAGE_LOAD_TIMEOUT
    self._script_timeout = script_timeout or self._DEFAULT_SCRIPT_TIMEOUT

    self._logger = logging.getLogger('airflow.task')

    driver = self._get_driver()
    super().__init__(driver)

  def _get_driver(self):
    
    if self._remote_hub:
      self._logger.info(f'driver mode: remote - {self._remote_hub}')
      return self._get_remote_driver()

    # 设置CFT driver和chrome路径
    self._set_cft()

    if self._undetected:
      # local udc driver
      return self._get_undetected_driver()

    # local standard driver
    return self._get_standard_driver()
  
  def _get_standard_driver(self):
    options, creator = self._get_options()
    service = Service(self._executable_path) 
    driver = webdriver.Chrome( service = service, options = options)
    # cdp must be set after driver created
    creator.set_cdp(driver)
    # set the timeout
    self._set_timeout(driver)
    return driver
  
  def _get_undetected_driver(self):
    options, creator = self._get_options()
    driver = udc_driver.Chrome( driver_executable_path = self._executable_path, options = options)
    # set the timeout
    self._set_timeout(driver)
    return driver
  
  def _get_remote_driver(self):
    self._undetected = False # udc don't support remote connection
    options, creator = self._get_options()
    driver = webdriver.Remote( command_executor = self._remote_hub, options = options)
    # cdp must be set after driver created
    creator.set_cdp(driver)
    # set the timeout
    self._set_timeout(driver)
    return driver
  
  def _get_options(self):
    creator = OptionsCreator(self._std_args,self._exp_args,self._cdp_args,self._sel_args,self._ext_args,self._undetected)
    options = creator.create()
    return options, creator
  
  def _set_timeout(self,driver):
    # set the page load timeout
    driver.set_page_load_timeout(self._page_load_timeout)
    driver.set_script_timeout(self._script_timeout)
  
  def _set_cft(self):
    if not self._executable_path:
      cft_result:dict = CFTInstaller.install()
      if cft_result.get('error'):
        raise Exception(f"安装CFT ChromeDriver失败: {cft_result.get('error')}")

      # CFT chrome和driver必须匹配使用
      driver_path = Path(cft_result.get('driver_path')).as_posix()
      chrome_path = Path(cft_result.get('chrome_path')).as_posix()
      self._executable_path = driver_path
      self._sel_args['binary_location'] = chrome_path
  



import sys,os,re,time
from blues_lib.sele.element.deco.InfoKeyDeco import InfoKeyDeco

from blues_lib.sele.waiter.Querier import Querier  
from blues_lib.type.file.File import File

class Shot():

  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver,5) 

  # == module 2: element shot == #
  @InfoKeyDeco('screenshot')
  def screenshot(self,loc_or_elem,file,parent_loc_or_elem=None,timeout=5):
    '''
    @description 指定元素截图,不支持base64格式
    @param {str} selector : css selector 
    @param {str} file 保存位置
    @returns {str} file_path
    '''
    file_path = file if file else self.__get_default_file()
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    shot_status = web_element.screenshot(file_path)
    return file_path if shot_status else ''

  def __get_default_file(self,prefix='elementshot'):
    dir_path = File.get_dir_path('screenshot') 
    file_name = File.get_file_name(prefix=prefix,extension='png')
    return os.path.join(dir_path,file_name)

  



from selenium.webdriver.remote.webelement import WebElement
from blues_lib.namespace.CrawlerName import CrawlerName
from blues_lib.crawler.page.PageCrawler import PageCrawler

class NextPageCrawler(PageCrawler):

  NAME = CrawlerName.Engine.NEXT_PAGE

  def _crawl_by_page(self)->list[dict]|None:
    rows:list[dict] = []
    # crawl the first page
    page_no:int = 1
    if page_rows:= self._crawl_next_page():
      rows.extend(page_rows)
    # crawl the next pages
    while self._has_next():
      page_no += 1
      self._next_page()
      if page_rows:= self._crawl_next_page():
        rows.extend(page_rows)
        
      if page_no >= self._max_page_no:
        break
      self._set_interval()
    return rows
    
  def _has_next(self)->bool:
    next_button:WebElement|None = self._browser.waiter.querier.query(self._page_selector)
    return next_button.is_enabled() if next_button else False

